# SPDX-FileCopyrightText: 2022 Helmholtz-Zentrum Dresden-Rossendorf (HZDR)
#
# SPDX-License-Identifier: GPL-3.0-or-later
import base64
import json
from datetime import datetime

import requests
from django.conf import settings
from django.contrib.auth.models import User
from django.shortcuts import reverse
from django.test import TestCase
from django.utils import timezone
from rest_framework import status
from rest_framework.viewsets import ModelViewSet

from .models import (
    DigitalObject,
    DigitalObjectAttributes,
    DigitalObjectRelation,
    Group,
    HeliportUser,
    Image,
    Project,
    Vocabulary,
)
from .serializers import types
from .utils import handles
from .utils.json_ld import LdContext, TargetDescription, reverse_context_mapping
from .utils.queries import digital_objects_by_type
from .views import HeliportModelViewSet, ProjectViewSet
from .vocabulary import SemanticDesktop


class WithHeliportUser(TestCase):
    def setUp(self):
        self.auth_user, is_new = User.objects.get_or_create(username="testuser")
        self.user, is_new = HeliportUser.objects.get_or_create(
            auth_user=self.auth_user, display_name="HELIPORTUSER321"
        )
        self.client.force_login(self.auth_user)


class SearchTest(WithHeliportUser):
    def test_access_search(self):
        response = self.client.get(reverse("core:search_base"))
        self.assertEqual(200, response.status_code)

    def test_search_valid_redirect_urls(self):
        response = self.client.get(reverse("core:search_base"))
        self.assertEqual(200, response.status_code)
        urls = response.context_data["search_urls"]
        self.assertIn(reverse("core:search"), urls)
        self.assertGreater(
            len(urls), 1, "modules other than heliport app are not searchable"
        )

        for url in urls:
            response = self.client.get(url)
            self.assertEqual(200, response.status_code)

    def test_group_findable(self):
        Group.objects.create(display_name="GROUP321")
        response = self.client.get(f'{reverse("core:search")}?q=321')
        self.assertEqual(200, response.status_code)
        self.assertContains(response, "GROUP321")

    def test_heliport_user_and_project_findable(self):
        Project.objects.create(owner=self.user, label="PROJECT321")
        response = self.client.get(f'{reverse("core:search")}?q=321')
        self.assertEqual(200, response.status_code)
        self.assertContains(response, "HELIPORTUSER321")
        self.assertContains(response, "PROJECT321")

    def test_digital_object_autocomplete(self):
        def d(obj):
            return dict(label=obj.label, pk=obj.digital_object_id)

        def q(term=None, **kwargs):
            if term is not None:
                kwargs["term"] = term
            response = self.client.get(
                f'{reverse("core:digital_object_autocomplete")}', kwargs
            )
            self.assertEqual(200, response.status_code)
            return response.json()

        type_prop = Vocabulary().type
        o1 = DigitalObject.objects.create(label="obj1", persistent_id="https://obj1")
        o1_2 = DigitalObject.objects.create(label="obj1_2")
        o2 = Image.objects.create(label="obj2")
        o1_other_user = DigitalObject.objects.create(
            label="obj1", owner=HeliportUser.objects.create()
        )
        DigitalObjectRelation.objects.create(subject=o1, predicate=type_prop, object=o2)
        DigitalObjectRelation.objects.create(
            subject=o1_2, predicate=type_prop, object=o2
        )
        DigitalObjectRelation.objects.create(subject=o2, predicate=type_prop, object=o1)

        self.assertEqualNoOrder([d(o1), d(o1_2)], q(q="obj1"))
        self.assertEqualNoOrder([d(o1), d(o1_2)], q(term="obj1"))
        self.assertEqualNoOrder([d(o1), d(o1_2)], q("Obj1", exact=False))
        self.assertEqual([d(o1)], q("Obj1", exact="case_insensitive"))
        self.assertEqual([], q("Obj1", exact=True))
        self.assertEqual([d(o1_2)], q("obj1_2", type=o2.digital_object_id))
        self.assertEqualNoOrder([d(o1), d(o1_2)], q(type=o2.digital_object_id))
        self.assertEqualNoOrder(
            [d(o1_2), d(o2)],
            self.client.get(
                f'{reverse("core:digital_object_autocomplete")}?q=2'
                f"&type={o1.digital_object_id}&type={o2.digital_object_id}"
            ).json(),
        )
        self.assertEqual([d(o2)], q(type="https://obj1"))
        self.assertEqual([d(o2)], q(type=SemanticDesktop.NFO.Image))
        self.assertEqual([d(o1)], q("https://obj1"))

        o1.deleted = timezone.now()
        o1.save()
        self.assertEqual([d(o1_2)], q("obj1"))

        for i in range(settings.HELIPORT_CORE_AUTOCOMPLETE_COUNT + 3):
            DigitalObject.objects.create(label=f"test_max_count{i}")
        self.assertEqual(
            settings.HELIPORT_CORE_AUTOCOMPLETE_COUNT, len(q("test_max_count"))
        )

        for o in DigitalObject.objects.exclude(
            digital_object_id__in=[o2.digital_object_id, o1_2.digital_object_id]
        ):
            o.deleted = timezone.now()
            o.save()
        self.assertEqualNoOrder([d(o1_2), d(o2)], q())

    def assertEqualNoOrder(self, list_goal, list_actual):
        missing = [o for o in list_goal if o not in list_actual]
        unexpected = [o for o in list_actual if o not in list_goal]
        self.assertEqual([], missing, "Some output is missing")
        self.assertEqual([], unexpected, "Some output is unexpected")


class HELIPORTTest(WithHeliportUser):
    def setUp(self):
        super().setUp()
        self.project, is_new = Project.objects.get_or_create(
            owner=self.user, label="xabac"
        )

    def test_project_graph(self):
        response = self.client.get(
            reverse("core:project_graph", kwargs={"project": self.project.pk})
        )
        self.assertEqual(200, response.status_code)
        self.assertContains(response, "xabac")

    def test_simple_api_docu(self):
        """mirroring the way it is described in projects api view"""

        response = self.client.get("/api-root/")
        self.assertEqual(200, response.status_code)
        urls = json.loads(response.content.decode())

        response = self.client.get(urls["projects"])
        self.assertEqual(200, response.status_code)
        response = self.client.get(f'{urls["projects"]}urls/')
        self.assertEqual(200, response.status_code)
        urls = json.loads(response.content.decode())

        self.assertIsInstance(urls, list)
        self.assertGreater(
            len(urls), 0, "there are no api urls other than the main heliport app api"
        )

    def test_importing_digital_objects(self):
        vocab = Vocabulary()
        import_project_url = reverse(
            "core:import_project", kwargs={"project": self.project.pk}
        )
        other_user = HeliportUser.objects.create(display_name="other_user")
        source_project = Project.objects.create(
            label="source_project", owner=other_user
        )
        obj1 = Image.objects.create(label="obj1")
        obj1.projects.add(source_project)
        obj1_id = obj1.digital_object_id
        obj2 = Image.objects.create(label="obj2")
        obj2.projects.add(source_project)
        obj2_id = obj2.digital_object_id
        obj3 = Image.objects.create(label="obj3")
        obj3.projects.add(source_project)
        obj3_id = obj3.digital_object_id
        import_request_data = {
            "import_objects": [str(obj1_id), str(obj2_id), str(obj3_id)],
            f"import-type-{obj1_id}": "import",
            f"import-type-{obj2_id}": "copy",
            f"import-type-{obj3_id}": "ignore",
        }

        self.assertNotContains(
            self.client.get(f"{import_project_url}?from={source_project.pk}"),
            "obj1",
            status_code=403,
        )
        self.client.post(
            f"{import_project_url}?from={source_project.pk}",
            folow=True,
            data=import_request_data,
        )
        self.assertFalse(self.project.digitalobject_set.exists())
        self.assertFalse(
            DigitalObjectRelation.objects.filter(predicate=vocab.is_version_of).exists()
        )
        self.assertEqual(3, source_project.digitalobject_set.count())

        source_project.co_owners.add(self.user)

        self.assertContains(
            self.client.get(f"{import_project_url}?from={source_project.pk}"), "obj1"
        )
        self.assertContains(
            self.client.get(f"{import_project_url}?from={source_project.pk}"), "Image"
        )
        import_response = self.client.post(
            f"{import_project_url}?from={source_project.pk}",
            follow=True,
            data=import_request_data,
        )
        self.assertEqual(200, import_response.status_code)

        self.assertEqual(3, source_project.digitalobject_set.count())
        self.assertEqual(2, self.project.digitalobject_set.count())
        obj2_copy_query = self.project.digitalobject_set.exclude(
            digital_object_id=obj1_id
        )
        self.assertEqual(1, obj2_copy_query.count())
        obj2_copy = obj2_copy_query.first().casted_or_self()
        self.assertNotEqual(obj2, obj2_copy)
        self.assertTrue(
            DigitalObjectRelation.objects.filter(
                subject=obj2_copy, predicate=vocab.is_version_of, object=obj2
            ).exists()
        )

    def test_copy_digital_object(self):
        vocab = Vocabulary()
        existing_objects = [obj.pk for obj in DigitalObject.objects.all()]
        obj = DigitalObject.objects.create(
            label="digital_object_to_copy", description="123", persistent_id="x"
        )
        DigitalObjectRelation.objects.create(
            subject=obj, predicate=vocab.type, object=vocab.Class
        )
        DigitalObjectAttributes.objects.create(
            subject=obj,
            predicate=vocab.primary_topic,
            value="x",
            special_heliport_role="r",
        )
        obj_copy = obj.create_copy()
        existing_objects.extend([obj_copy.pk, obj.pk])
        self.assertNotEqual(obj, obj_copy)
        self.assertEqual("digital_object_to_copy", obj_copy.label)
        self.assertEqual("123", obj_copy.description)
        self.assertIsNone(obj_copy.persistent_id)
        self.assertIsNone(obj_copy.generated_persistent_id)
        self.assertTrue(
            DigitalObjectRelation.objects.filter(
                subject=obj_copy, predicate=vocab.is_version_of, object=obj
            ).exists()
        )
        self.assertTrue(
            DigitalObjectRelation.objects.filter(
                subject=obj_copy, predicate=vocab.type, object=vocab.Class
            ).exists()
        )
        self.assertTrue(
            DigitalObjectAttributes.objects.filter(
                subject=obj_copy,
                predicate=vocab.primary_topic,
                value="x",
                special_heliport_role="r",
            ).exists()
        )

        im = Image.objects.create(filename="fn", label="test_image")
        im_copy = im.create_copy()
        self.assertNotEqual(im, im_copy)
        self.assertEqual("fn", im_copy.filename)
        self.assertEqual("test_image", im_copy.label)
        existing_objects.extend([im.digital_object_id, im_copy.digital_object_id])

        self.assertFalse(
            DigitalObject.objects.exclude(
                digital_object_id__in=existing_objects
            ).exists()
        )
        for o in existing_objects:
            self.assertTrue(DigitalObject.objects.get(digital_object_id=o))


class APITest(WithHeliportUser):
    def test_digital_object_api(self):
        # api should be available at this path even when not logged in
        self.client.logout()
        result = self.client.get("/api/digital_object_api", follow=True)
        self.assertEqual(200, result.status_code)

    def test_deletion(self):
        self.assertFalse(Project.objects.exists())
        new_project = Project.objects.create(owner=self.user, label="to delete project")
        self.assertIsNone(new_project.deleted)
        response = self.client.delete(f"/api/projects/{new_project.pk}/")
        self.assertEqual(status.HTTP_204_NO_CONTENT, response.status_code)
        self.assertTrue(Project.objects.exists())
        new_project.refresh_from_db()
        self.assertIsNotNone(new_project.deleted)

        # ensure all DigitalObject ViewSets are HELIPORT ViewSets like project ViewSet
        view_set_to_model_path = ["serializer_class", "Meta", "model"]
        self.assertEqual(
            Project, self.getattr_path(ProjectViewSet, view_set_to_model_path)
        )
        self.assertTrue(issubclass(Project, DigitalObject))
        invalid = []
        valid = []
        other_count = 0
        for model_view_set in self.subclasses_recursive(ModelViewSet):
            the_model = self.getattr_path(model_view_set, view_set_to_model_path)
            if the_model is not None and issubclass(the_model, DigitalObject):
                description = f"{model_view_set.__name__} of {the_model.__name__}"
                if issubclass(model_view_set, HeliportModelViewSet):
                    valid.append(description)
                else:
                    invalid.append(description)
            else:
                other_count += 1

        self.assertEqual(
            [],
            invalid,
            f"{len(invalid)}/{len(invalid) + len(valid)} (ignored {other_count}) are invalid\n"
            f"Digital Object ModelViewSets must be subclasses of {HeliportModelViewSet.__name__}",
        )

    @staticmethod
    def getattr_path(obj, name_path, default=None):
        for name in name_path:
            obj = getattr(obj, name, default)
        return obj

    def subclasses_recursive(self, the_class):
        yield the_class
        for subclass in the_class.__subclasses__():
            yield from self.subclasses_recursive(subclass)

    def test_token_auth(self):
        # Note that these absolute urls may be used in HELIPORT:
        token_response = self.client.post("/api/auth/login/")
        token1 = token_response.json()["token"]
        token_response = self.client.post("/api/auth/login/")
        token2 = token_response.json()["token"]
        self.assertEqual(status.HTTP_200_OK, token_response.status_code)
        self.assertEqual(
            status.HTTP_204_NO_CONTENT,
            self.client.post(
                "/api/auth/logout/", HTTP_AUTHORIZATION=f"Token {token1}"
            ).status_code,
        )
        self.assertEqual(
            status.HTTP_401_UNAUTHORIZED,
            self.client.post(
                "/api/auth/logout/", HTTP_AUTHORIZATION=f"Token {token1}"
            ).status_code,
        )
        self.assertEqual(
            status.HTTP_204_NO_CONTENT,
            self.client.post(
                "/api/auth/logoutall/", HTTP_AUTHORIZATION=f"Token {token2}"
            ).status_code,
        )


class HandleServerInteractionTest(TestCase):
    b64_prefix = "data:application/json;charset=utf-8;base64,"
    json_header = {"Content-Type": "application/json"}
    auth_header = {
        "Content-Type": "application/json",
        "Authorization": 'Handle clientCert="true"',
    }
    auth_cert = (
        settings.HELIPORT_HANDLE_CERTIFICATE_PATH,
        settings.HELIPORT_HANDLE_PRIVATE_KEY_PATH,
    )

    class ResponseObj:
        def __init__(self, status=None, json_data=None, ok=True):
            self.status_code = status
            self.json_data = json_data
            self.ok = ok

        def json(self):
            return self.json_data

    class HandleServerMock(handles.HandleServer):
        def __init__(self, response_list):
            self.request_list = []
            self.response_list = response_list
            self.status = -1

        def do_request(self, **kwargs):
            self.request_list.append(kwargs)
            self.status += 1
            return self.response_list[self.status]

    def test_json_data_url(self):
        date_created = datetime(2022, 1, 2, 13, 4)
        data_url = handles.generate_handle_json_data_url(
            "test_handle_name",
            None,
            "test/category",
            "https://www.example.com",
            date_created,
        )
        self.assertTrue(data_url.startswith(self.b64_prefix))
        b64_data = data_url[len(self.b64_prefix) :]
        json_data = json.loads(base64.b64decode(b64_data).decode("utf-8"))
        self.assertEqual("test_handle_name", json_data.get("id"))
        self.assertEqual("2022/01/02, 13:04", json_data.get("created_on"))
        self.assertEqual("https://www.example.com", json_data.get("url"))

    def test_json_data_url_local_url(self):
        date_created = datetime(2022, 1, 2, 13, 4)
        data_url_no_local = handles.generate_handle_json_data_url(
            "test_handle_name",
            "https://www.example.com",
            "test/category",
            "https://www.example.com",
            date_created,
        )
        data_url_local = handles.generate_handle_json_data_url(
            "test_handle_name",
            "https://localhost",
            "test/category",
            "https://www.example.com",
            date_created,
        )
        b64_data_no_local = data_url_no_local[len(self.b64_prefix) :]
        b64_data_local = data_url_local[len(self.b64_prefix) :]
        json_data_no_local = json.loads(
            base64.b64decode(b64_data_no_local).decode("utf-8")
        )
        json_data_local = json.loads(base64.b64decode(b64_data_local).decode("utf-8"))
        self.assertIsNone(json_data_no_local.get("local_url"))
        self.assertEqual("https://localhost", json_data_local.get("local_url"))

    def test_handle_registering(self):
        response = self.ResponseObj(200)
        server = self.HandleServerMock([response])
        self.assertEqual(
            response,
            handles.register_handle("test_handle", "https://www.example.com", server),
        )
        self.assertEqual(
            [
                dict(
                    url=f"{settings.HELIPORT_HANDLE_API_URL}/{settings.HELIPORT_HANDLE_PREFIX}/test_handle?overwrite=false",
                    method=requests.put,
                    headers=self.auth_header,
                    cert=self.auth_cert,
                    json={
                        "index": 1,
                        "type": "URL",
                        "data": {
                            "format": "string",
                            "value": "https://www.example.com",
                        },
                    },
                )
            ],
            server.request_list,
        )

    def test_handle_registering_handle_already_exists(self):
        response = self.ResponseObj(200)

        server = self.HandleServerMock([self.ResponseObj(409), response])
        self.assertEqual(
            response,
            handles.register_handle("test_handle2", "https://www.example2.com", server),
        )
        self.assertEqual(
            f"{settings.HELIPORT_HANDLE_API_URL}/{settings.HELIPORT_HANDLE_PREFIX}/test_handle2?overwrite=false",
            server.request_list[0]["url"],
        )
        self.assertEqual(
            f"{settings.HELIPORT_HANDLE_API_URL}/{settings.HELIPORT_HANDLE_PREFIX}/?mintNewSuffix=true",
            server.request_list[1]["url"],
        )
        self.assertEqual(requests.put, server.request_list[1]["method"])
        self.assertEqual(
            {
                "index": 1,
                "type": "URL",
                "data": {
                    "format": "string",
                    "value": "https://www.example2.com",
                },
            },
            server.request_list[1]["json"],
        )
        self.assertEqual(2, len(server.request_list))

    def test_extract_handle(self):
        self.assertEqual(
            "https://hdl.handle.net/123/456",
            handles.get_handle_from_response(
                self.ResponseObj(json_data={"handle": "123/456"})
            ),
        )

    def test_get_key_values(self):
        server = self.HandleServerMock(
            [self.ResponseObj(ok=False), self.ResponseObj(ok=False)]
        )
        self.assertEqual(
            [], handles.get_key_values("https://hdl.handle.net/123/456", server)
        )
        self.assertEqual(
            [],
            handles.get_key_values(
                f"https://hdl.handle.net/{settings.HELIPORT_HANDLE_PREFIX}/555", server
            ),
        )
        self.assertEqual(
            [
                dict(
                    method=requests.get,
                    url="https://hdl.handle.net/api/handles/123/456",
                    headers=self.json_header,
                ),
                dict(
                    method=requests.get,
                    url=f"{settings.HELIPORT_HANDLE_API_URL}/{settings.HELIPORT_HANDLE_PREFIX}/555",
                    headers=self.json_header,
                ),
            ],
            server.request_list,
        )

    def test_get_key_values_real_request(self):
        self.assertEqual(
            [
                ("INSTITUTION", "Helmholtz-Zentrum Dresden-Rossendorf"),
                ("AUTHOR", "Martin Voigt (FWCC) - 141575"),
                (
                    "DESCRIPTION",
                    "This is a handle with entries of not supported types.",
                ),
            ],
            handles.get_key_values("20.500.12865/custom_properties_handle"),
        )
        print(
            "Making real request to handle proxy, testing if handle server is reachable"
        )

    def test_url_not_in_key_values(self):
        server = self.HandleServerMock(
            [
                self.ResponseObj(
                    json_data={
                        "responseCode": 1,
                        "handle": "123/456",
                        "values": [
                            {"type": "t1", "data": {"value": {"foo": 1}}},
                            {"type": "URL", "data": {"value": "v2"}},
                        ],
                    }
                )
            ]
        )
        self.assertEqual(
            [("t1", "{'foo': 1}")], handles.get_key_values("123/456", server)
        )

    def test_set_key_values(self):
        json_url = json.dumps(
            {"url": f"{settings.HELIPORT_HANDLE_DIGITAL_OBJECT_HOST}/objects/42"}
        ).encode()
        b64_url = self.b64_prefix + base64.b64encode(json_url).decode()
        server = self.HandleServerMock(
            [
                self.ResponseObj(
                    json_data={
                        "responseCode": 1,
                        "handle": f"{settings.HELIPORT_HANDLE_PREFIX}/456",
                        "values": [
                            {"type": "t1"},
                            {
                                "type": "URL",
                                "data": {"value": b64_url},
                            },
                            {"type": "t3"},
                        ],
                    }
                ),
                self.ResponseObj(ok=True),
            ]
        )

        self.assertTrue(
            handles.set_key_values(
                f"https://hdl.handle.net/{settings.HELIPORT_HANDLE_PREFIX}/456",
                [
                    ("ABC", "DEF"),
                    ("https://hdl.handle.net/UVW", "XYZ"),
                    ("https://doi.org/123", "456"),
                ],
                server,
                simulate_register_handles_in_tests=True,
            )
        )

        self.assertEqual(
            [
                {
                    "url": f"{settings.HELIPORT_HANDLE_API_URL}/{settings.HELIPORT_HANDLE_PREFIX}/456",
                    "method": requests.get,
                    "headers": self.json_header,
                },
                {
                    "url": f"{settings.HELIPORT_HANDLE_API_URL}/{settings.HELIPORT_HANDLE_PREFIX}/456",
                    "method": requests.put,
                    "headers": self.auth_header,
                    "cert": self.auth_cert,
                    "json": [
                        {
                            "index": 1,
                            "type": "URL",
                            "data": {
                                "format": "string",
                                "value": b64_url,
                            },
                        },
                        {
                            "index": 2,
                            "type": "ABC",
                            "data": {"format": "string", "value": "DEF"},
                        },
                        {
                            "index": 3,
                            "type": "UVW",
                            "data": {"format": "string", "value": "XYZ"},
                        },
                        {
                            "index": 4,
                            "type": "123",
                            "data": {"format": "string", "value": "456"},
                        },
                    ],
                },
            ],
            server.request_list,
        )


class CoreVocabularyTest(TestCase):
    def test_vocabularies_exist_and_are_property_or_class(self):
        vocab = Vocabulary()
        for voc in dir(vocab):
            if not voc.startswith("_"):
                obj = getattr(vocab, voc)
                self.assertIsNotNone(obj.label)
                property_or_class = vocab.Class in types(
                    obj
                ) or vocab.Property in types(obj)
                if not property_or_class:
                    self.fail(
                        f"{obj} is not a property or class. types are: {types(obj)}"
                    )

    def test_query_by_type(self):
        vocab = Vocabulary()
        p1 = Project.objects.create()
        p2 = Project.objects.create()
        i1o = Image.objects.create()
        i2o = Image.objects.create()
        t = DigitalObject.objects.create()
        i1 = DigitalObject.objects.get(digital_object_id=i1o.digital_object_id)
        i2 = DigitalObject.objects.get(digital_object_id=i2o.digital_object_id)

        self.assertEqual(
            [{p1, p2}], list(map(set, digital_objects_by_type(vocab.Project)))
        )
        DigitalObjectRelation.objects.create(
            subject=i1, predicate=vocab.type, object=vocab.Project
        )
        self.assertEqual(
            [{i1}, {p1, p2}], list(map(set, digital_objects_by_type(vocab.Project)))
        )
        self.assertEqual(
            [{i1}, {p1, p2}],
            list(map(set, digital_objects_by_type(vocab.Project.digital_object_id))),
        )
        self.assertEqual(
            [{i1}, {p1, p2}],
            list(
                map(set, digital_objects_by_type(str(vocab.Project.digital_object_id)))
            ),
        )
        self.assertEqual(
            [{i1}, {p1, p2}],
            list(map(set, digital_objects_by_type(vocab.Project.persistent_id))),
        )
        self.assertEqual([], digital_objects_by_type("https://abc"))

        t.persistent_id = vocab.Project.persistent_id
        t.save()
        DigitalObjectRelation.objects.create(subject=i2, predicate=vocab.type, object=t)

        self.assertEqual(
            [{i1}, {p1, p2}], list(map(set, digital_objects_by_type(vocab.Project)))
        )
        self.assertEqual(
            [{i1, i2}, {p1, p2}],
            list(map(set, digital_objects_by_type(vocab.Project.persistent_id))),
        )
        self.assertEqual(
            [{i1o, i2o}],
            list(map(set, digital_objects_by_type(SemanticDesktop.NFO.Image))),
        )


class LdJsonContextTest(TestCase):
    def test_basic_set(self):
        context = LdContext()
        a = dict()
        b = dict()
        context.tentative_set(a, "x1", 1, "https://x")
        context.tentative_set(a, "y1", 2, "https://y")
        context.tentative_set(a, "y2", 3, "https://y")
        context.tentative_set(b, "x1", 4, "https://x")
        context.tentative_set(b, "y1", 5, "https://y")
        context.tentative_set(b, "y2", 6, "https://y")
        self.assertEqual(dict(x1=1, y1=2, y2=3), a)
        self.assertEqual(dict(x1=4, y1=5, y2=6), b)
        self.assertEqual(
            dict(x1="https://x", y1="https://y", y2="https://y"),
            context.get_context_dict(),
        )

    def test_unique_keys(self):
        context = LdContext()
        a = dict()
        b = dict()
        context.tentative_set(a, "x", 1, "https://x0")
        context.tentative_set(a, "x", 2, "https://x1")
        context.tentative_set(b, "x", 3, "https://x1")
        context.tentative_set(b, "x", 4, "https://x0")
        context.tentative_set(a, "x", 5, "https://x1")

        # note https://x1 comes alphabetically before https://x2
        self.assertEqual(dict(x=1, x_1=5), a)
        self.assertEqual(dict(x=4, x_1=3), b)
        self.assertEqual(
            dict(x="https://x0", x_1="https://x1"), context.get_context_dict()
        )

    def test_rename_keys(self):
        context = LdContext()
        a = dict()
        b = dict()
        context.tentative_set(a, "x", 1, "https://x1")
        context.tentative_set(b, "x", 2, "https://x1")
        self.assertEqual(dict(x=1), a)
        self.assertEqual(dict(x=2), b)
        # renames old x to x1 because https://x0 comes alphabetically before https://x1:
        context.tentative_set(b, "x", 3, "https://x0")
        context.tentative_set(a, "x", 4, "https://x0")
        context.tentative_set(a, "x", 5, "https://x1")

        self.assertEqual(dict(x=4, x_1=5), a)
        self.assertEqual(dict(x=3, x_1=2), b)
        self.assertEqual(
            dict(x="https://x0", x_1="https://x1"), context.get_context_dict()
        )

    def test_chained_renaming(self):
        context = LdContext()
        a = dict()
        context.tentative_set(a, "x_2", 3, "https://x3")
        context.tentative_set(a, "x", 2, "https://x2")
        # renames x -> x1 because "x1" < "x2":
        context.tentative_set(a, "x", 1, "https://x1")
        # renames x3 -> x31; x1 -> x2; x -> x1 because "x0" < "x1" < "x2" < "x3":
        context.tentative_set(a, "x", 0, "https://x0")

        self.assertEqual(dict(x=0, x_1=1, x_2=2, x_2_1=3), a)

    def test_add_to_end_of_chain(self):
        context = LdContext()
        a = dict()
        context.tentative_set(a, "x_2", 3, "https://x3")
        context.tentative_set(a, "x", 0, "https://x0")
        # adds x as x1
        context.tentative_set(a, "x", 1, "https://x1")
        # add x as x2 and renames x2 -> x21
        context.tentative_set(a, "x", 2, "https://x2")

        self.assertEqual(dict(x=0, x_1=1, x_2=2, x_2_1=3), a)
        self.assertEqual(
            dict(
                x="https://x0", x_1="https://x1", x_2="https://x2", x_2_1="https://x3"
            ),
            context.get_context_dict(),
        )

    def test_match_context(self):
        target1 = TargetDescription("a", "https://a", 1)
        self.assertEqual(set(), set(reverse_context_mapping([], dict())))
        self.assertEqual(set(), set(reverse_context_mapping([target1], dict())))
        self.assertEqual(set(), set(reverse_context_mapping([], dict(a=1))))

        self.assertEqual(
            {(1, -1), (2, -2), (3, -3)},
            set(
                reverse_context_mapping(
                    [
                        TargetDescription("a", "https://a", 1),
                        TargetDescription("b", "https://b", 2),
                        TargetDescription("c_1", "https://c", 3),
                    ],
                    dict(a=-1, b_2=-2, c_1_3=-3),
                )
            ),
        )

        self.assertEqual(
            {(1, -1), (2, -2), (3, -3), (4, -4)},
            set(
                reverse_context_mapping(
                    [
                        TargetDescription("a", "https://a0", 1),
                        TargetDescription("a", "https://a1", 2),
                        TargetDescription("a_1", "https://a2", 3),
                        TargetDescription("a", "https://a3", 4),
                    ],
                    dict(a=-1, a_1=-2, a_1_1=-3, a_2=-4),
                )
            ),
        )

        self.assertEqual(
            {(1, -1), (2, -2), (3, -3)},
            set(
                reverse_context_mapping(
                    [
                        TargetDescription("a", "https://a2", 3),
                        TargetDescription("a", "https://a1", 2),
                        TargetDescription("a_1", "https://a0", 1),
                    ],
                    dict(a=-2, a_2=-3, a_1=-1),
                )
            ),
        )

    def test_reverse_compatible_with_generation(self):
        context = LdContext()
        a = dict()
        b = dict()

        context.tentative_set(b, "a", 0, "https://a0")
        context.tentative_set(a, "a", 1, "https://a1")
        context.tentative_set(a, "a_1", 2, "https://a2")
        context.tentative_set(a, "a_1", 3, "https://a3")
        context.tentative_set(a, "b", 4, "https://b")

        self.assertEqual(dict(a=0), b)
        self.assertEqual(dict(a_1=1, a_1_1=2, a_1_2=3, b=4), a)

        self.assertEqual(
            {(-1, 1), (-2, 2), (-3, 3), (dict, 4)},
            set(
                reverse_context_mapping(
                    [
                        TargetDescription("a", "https://a1", -1),
                        TargetDescription("a_1", "https://a2", -2),
                        TargetDescription("a_1", "https://a3", -3),
                        TargetDescription("b", "https://b", dict),
                    ],
                    a,
                )
            ),
        )
