# SPDX-FileCopyrightText: 2022 Helmholtz-Zentrum Dresden-Rossendorf (HZDR)
#
# SPDX-License-Identifier: GPL-3.0-or-later
from django.conf import settings
from django.urls import include, path, re_path
from django.views.decorators.cache import cache_page
from django.views.generic import RedirectView
from rest_framework import routers

from .views import (
    ContributionViewSet,
    CreateEmptyProjectView,
    CreateProjectView,
    DeleteAllTokensView,
    DigitalObjectAPI,
    DigitalObjectAutocompleteView,
    DigitalObjectQRView,
    FaviconView,
    GroupViewSet,
    ImagesView,
    ImageView,
    ImportFromProjectView,
    IndexView,
    LandingPageByPersistentIDView,
    LandingPageView,
    LoginInfoUpdateView,
    LoginInfoView,
    LoginNamedTokenView,
    LoginView,
    LogoutTokenView,
    LogoutView,
    MaintenanceMessageView,
    ManualUserDownloadView,
    ProjectGraphView,
    ProjectListView,
    ProjectTimelineView,
    ProjectUpdateView,
    ProjectView,
    ProjectViewSet,
    PublicLandingPageView,
    RedocView,
    RegisterView,
    RemoteServerLoginView,
    RemoteServerLogoutView,
    SchemaView,
    SearchBaseView,
    SearchView,
    SubprojectsView,
    TagAutocompleteView,
    TagJsonView,
    TagsView,
    TakePictureView,
    TokenViewSet,
    UserSettingsView,
    UserViewSet,
    get_project_list,
    get_users_autocomplete,
)

app_name = "core"
urlpatterns = [
    re_path(
        "about/(?P<path>.*)",
        RedirectView.as_view(url=f"{settings.HELIPORT_CORE_ABOUT_BASE_URL}/%(path)s"),
        name="about",
    ),
    path("login/", LoginView.as_view(), name="login"),
    path("register/", RegisterView.as_view(), name="register"),
    path("logout/", LogoutView.as_view(), name="logout"),
    path("user/settings/", UserSettingsView.as_view(), name="user_settings"),
    path(
        "user/logout_all_tokens/",
        DeleteAllTokensView.as_view(),
        name="logout_all_tokens",
    ),
    path("user/logins/list/", LoginInfoView.as_view(), name="user_login_list"),
    path(
        "user/logins/update/<int:pk>/",
        LoginInfoUpdateView.as_view(),
        name="user_login_update",
    ),
    path(
        "user/logins/remote_server_login/<int:pk>/",
        RemoteServerLoginView.as_view(),
        name="remote_server_login",
    ),
    path(
        "user/logins/remote_server_logout/<int:pk>/",
        RemoteServerLogoutView.as_view(),
        name="remote_server_logout",
    ),
    path(
        "digital_object_autocomplete/",
        DigitalObjectAutocompleteView.as_view(),
        name="digital_object_autocomplete",
    ),
    path(
        "public/object/H<int:pk>/",
        PublicLandingPageView.as_view(),
        name="public_landing_page",
    ),
    path("object/<int:pk>/", LandingPageView.as_view(), name="landing_page"),
    # required by xml serialization:
    path("object/H<int:pk>/", LandingPageView.as_view(), name="prefixed_landing_page"),
    path(
        "object/<int:pk>/qr/", DigitalObjectQRView.as_view(), name="digital_object_qr"
    ),
    path(
        "object/",
        LandingPageByPersistentIDView.as_view(),
        name="digital_object_by_persistent_id",
    ),
    path("project/<int:project>/images/", ImagesView.as_view(), name="image_list"),
    path(
        "project/<int:project>/images/take/",
        TakePictureView.as_view(),
        name="take_image",
    ),
    path(
        "images/<int:pk>/data/",
        cache_page(4 * 24 * 60 * 60, cache="image_cache")(ImageView.as_view()),
        name="image_data",
    ),
    path("project/<int:project>/tags/", TagsView.as_view(), name="tag_list"),
    path("project/<int:project>/tags/<int:pk>/", TagsView.as_view(), name="tag_update"),
    path(
        "project/<int:project>/tags/json/",
        TagJsonView.as_view(),
        name="tag_json",
    ),
    path(
        "project/<int:project>/tags/autocomplete/",
        TagAutocompleteView.as_view(),
        name="tag_autocomplete",
    ),
    path(
        "project/<int:project>/subprojects/",
        SubprojectsView.as_view(),
        name="subprojects",
    ),
    path(
        "project/<int:project>/import/",
        ImportFromProjectView.as_view(),
        name="import_project",
    ),
    path("project/list/", ProjectListView.as_view(), name="project_list"),
    path(
        "project/create/empty/",
        CreateEmptyProjectView.as_view(),
        name="project_create_empty",
    ),
    path("project/create/", CreateProjectView.as_view(), name="project_create"),
    path(
        "project/<int:project>/graph/", ProjectGraphView.as_view(), name="project_graph"
    ),
    path(
        "project/<int:project>/timeline/",
        ProjectTimelineView.as_view(),
        name="project_timeline",
    ),
    path("project/<int:project>/", ProjectView.as_view(), name="project"),
    path(
        "project/update/<int:project>/",
        ProjectUpdateView.as_view(),
        name="project_update",
    ),
    path("manually/update/user/database/", ManualUserDownloadView.as_view()),
    path("", IndexView.as_view(), name="index"),
    path("project/<int:project>/autocomplete_user/", get_users_autocomplete),
    path(
        "project/dropdown_project_list/", get_project_list, name="dropdown_project_list"
    ),
    path("search_base/", SearchBaseView.as_view(), name="search_base"),
    path("search/", SearchView.as_view(), name="search"),
    path("favicon.ico", FaviconView.as_view()),
    path(
        "maintenance_message",
        MaintenanceMessageView.as_view(),
        name="maintenance_message",
    ),
]

# for REST API:
router = routers.DefaultRouter()
router.register(r"projects", ProjectViewSet, basename="projects")
router.register(r"users", UserViewSet, basename="users")
router.register(r"groups", GroupViewSet, basename="groups")
router.register(r"contributions", ContributionViewSet, basename="contributions")
router.register(r"tokens", TokenViewSet, basename="tokens")
router.register(r"digital_object_api", DigitalObjectAPI, basename="digital_object_api")

urlpatterns += [
    path("api-root/", router.get_api_root_view(), name="api_root"),
    path("api/", include(router.urls)),
    path("api/auth/delete_token/", LogoutTokenView.as_view(), name="delete_token"),
    path("api/auth/login_named/", LoginNamedTokenView.as_view(), name="login_named"),
    path("api/auth/", include("knox.urls")),
    path("openapi/", SchemaView, name="openapi-schema"),
    path("redoc/", RedocView.as_view(), name="redoc"),
]
