# SPDX-FileCopyrightText: 2022 Helmholtz-Zentrum Dresden-Rossendorf (HZDR)
#
# SPDX-License-Identifier: GPL-3.0-or-later
import typing
from contextlib import contextmanager
from itertools import chain

from django.conf import settings
from django.db.models import QuerySet
from django.shortcuts import get_object_or_404

from heliport.core.attribute_description import ConstantAttribute
from heliport.core.models import (
    DigitalObject,
    DigitalObjectRelation,
    Group,
    HeliportUser,
    Project,
    Vocabulary,
    assert_relation,
    digital_object_form_label_and_pk,
)
from heliport.core.serializers import static_attributes
from heliport.core.user_logic.user_information import AuthenticationBackendManager
from heliport.core.utils.IDs import generate_project_id


def create_empty_project(request):
    auth_user = request.user
    heliport_user = get_object_or_404(HeliportUser, auth_user=auth_user)
    project = Project(owner=heliport_user)

    group_name = AuthenticationBackendManager().get_group(auth_user)
    if group_name is not None:
        try:
            project.group = Group.objects.filter(display_name=group_name).get()
        except Group.DoesNotExist:
            new_group = Group(display_name=group_name)
            new_group.save()
            project.group = new_group

    project.save()
    project.hzdr_id = generate_project_id(auth_user, project.group_id, project.created)
    group_str = ""
    if project.group is not None:
        group_str = f"/{project.group}"
    project.category_str = f"{settings.HELIPORT_CORE_HZDR_ID_PREFIX}/Projects/{project.created.year}{group_str}/Project"
    project.save()
    project.register_handle_if_applicable(request)

    return project


@contextmanager
def at_least_one_department(default_name="default_group"):
    g = Group.objects.first()
    if not g:
        g, is_new = Group.objects.get_or_create(display_name=default_name)
    yield g


def get_or_create_from_label(
    label, pk, request, c_category=None, c_type=None, c_project=None, default=None
):
    label_object = digital_object_form_label_and_pk(label, pk)
    if label_object is None and not label:
        return default
    if label_object is None:
        label_object = DigitalObject(label=label)
        if c_category is None:
            c_category = settings.DEFAULT_OBJECT_NAMESPACE
        label_object.category_str = c_category
        label_object.save()
        if c_project is not None:
            label_object.projects.add(c_project)
        if c_type is not None:
            assert_relation(label_object, Vocabulary().type, c_type)
        label_object.register_handle_if_applicable(request)
    return label_object


def digital_objects_by_type(
    the_type: typing.Union[int, str, DigitalObject]
) -> typing.List[QuerySet]:
    """
    Returns a list of QuerySets of objects of the_type.
    Each QuerySet is either a filter() of digital objects or all() objects of a model.

    :param the_type: DigitalObject that is the type. int or decimal str queries digital object by digital_object_id ; other string queries by persistent_id
    """
    result = []
    type_attr = Vocabulary().type

    if isinstance(the_type, str) and the_type.isdecimal():
        the_type = int(the_type)

    if isinstance(the_type, int):
        the_type = DigitalObject.objects.get(digital_object_id=the_type)

    if isinstance(the_type, str):
        type_objects = DigitalObject.objects.filter(persistent_id=the_type)
    else:
        type_objects = [the_type]

    dynamic_results = DigitalObject.objects.filter(
        relations__in=DigitalObjectRelation.objects.filter(
            predicate=type_attr, object__in=type_objects
        )
    )
    if dynamic_results.exists():
        result.append(dynamic_results)

    for a_class, attributes in static_attributes.items():
        if issubclass(a_class, DigitalObject) and any(
            isinstance(attr, ConstantAttribute)
            and attr.attribute == type_attr
            and (attr.value in type_objects or str(attr.value) == str(the_type))
            for attr in attributes()
        ):
            result.append(a_class.objects.all())

    return result


class ListOfQuerySets:
    def __init__(self, list_of_query_sets: typing.List[QuerySet] = None):
        self.data = list_of_query_sets
        if self.data is None:
            self.data = []

    def extend(self, list_of_query_sets: typing.List[QuerySet]):
        if isinstance(list_of_query_sets, ListOfQuerySets):
            list_of_query_sets = list_of_query_sets.data
        self.data.extend(list_of_query_sets)

    def all_digital_objects(self):
        self.data = [DigitalObject.objects.exclude(label=12334555)]

    def filter(self, *args, **kwargs):
        self.data = [q.filter(*args, **kwargs) for q in self.data]

    def as_list(self, max_count=None):
        if max_count is None:
            return list(set(chain(*self.data)))
        else:
            objects = set()
            for q in self.data:
                objects.update(q[:max_count])
                if len(objects) > max_count:
                    break
            return list(objects)[:max_count]
