# SPDX-FileCopyrightText: 2022 Helmholtz-Zentrum Dresden-Rossendorf (HZDR)
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""
This module contains `django class based views <https://docs.djangoproject.com/en/4.1/topics/class-based-views/>`_
to handle incoming http requests. Each ``View`` is mapped to a URL in :mod:`heliport.core.urls`.

Each django app can have its own views. This module contains views for all main heliport functionality.
"""

import http.client
import io
import logging
from collections import OrderedDict
from datetime import datetime, timedelta

import django_filters.rest_framework
import knox.views
import qrcode
from django.conf import settings
from django.contrib import messages
from django.contrib.auth import authenticate, logout, user_logged_out
from django.contrib.auth.mixins import UserPassesTestMixin
from django.contrib.auth.models import User
from django.contrib.staticfiles import finders
from django.db.models import F, Q
from django.db.models.functions import Lower
from django.http import JsonResponse
from django.shortcuts import Http404, HttpResponse, get_object_or_404, redirect, render
from django.urls import reverse
from django.utils import timezone
from django.views.generic import (
    CreateView,
    DetailView,
    ListView,
    RedirectView,
    TemplateView,
    UpdateView,
    View,
)
from knox import models as knox_models
from rest_framework import filters as rest_filters
from rest_framework import permissions as rest_permissions
from rest_framework import status, viewsets
from rest_framework.decorators import action
from rest_framework.renderers import BrowsableAPIRenderer
from rest_framework.schemas import get_schema_view
from rest_framework.settings import api_settings
from rest_framework.views import APIView
from rest_framework.views import Response as RestResponse

import heliport.core.tasks
from heliport.core.utils.version import heliport_version_string
from heliport.utils.colors import pick_color_not_in
from heliport.utils.form_description import FormDescription
from heliport.utils.string_tools import remove_prefix

from .app_interaction import (
    generate_graph,
    get_api_roots,
    get_project_create_urls,
    get_search_urls,
)
from .custom_views import (
    HeliportLoginRequiredMixin,
    HeliportObjectListView,
    HeliportObjectMixin,
    HeliportProjectMixin,
)
from .models import (
    Contribution,
    DigitalObject,
    DigitalObjectRelation,
    Group,
    HeliportGroup,
    HeliportUser,
    Image,
    LoginInfo,
    NamedToken,
    Project,
    Tag,
    TimelineConfiguration,
    Vocabulary,
    assert_not_relation,
    assert_relation,
    digital_object_form_label_and_pk,
)
from .permissions import (
    HeliportObjectPermission,
    HeliportProjectPermission,
    StaffOrReadOnly,
    has_read_permission,
    has_write_permission,
    is_object_member,
    projects_of,
)
from .serializers import (
    ContributionSerializer,
    DigitalObjectAPISerializer,
    GroupSerializer,
    HtmlRenderer,
    LandingPageRenderer,
    ProjectSerializer,
    TokenSerializer,
    UserSerializer,
    get_all_attributes,
    get_attributes_by_category,
    get_unique_renderers,
    select_renderer,
    types,
)
from .user_logic.remote_login import delete_ssh_key, deploy_ssh_key
from .user_logic.user_information import AuthenticationBackendManager
from .utils.json_ld import LdContext
from .utils.queries import (
    ListOfQuerySets,
    at_least_one_department,
    create_empty_project,
    digital_objects_by_type,
)

logger = logging.getLogger(__name__)


class LoginView(TemplateView):
    """Handle main heliport login. Includes showing login screen."""

    #: HTML template rendered by this View
    template_name = "core/user/heliport_login.html"

    def get_context_data(self, **kwargs):
        """Called by django to get context when rendering template."""
        context = super().get_context_data(**kwargs)
        context["redirect_to"] = self.request.GET.get("next", None)
        context["enable_register"] = settings.AUTHENTICATION_BACKEND_USED == "basic"
        context["is_login_page"] = True
        return context

    def post(self, request):
        """Called by django on http post.

        Show that login failed or log in user. Fetch user information via a concrete
        :class:`heliport.core.user_logic.user_information_managers.managers.UserInformationManager`.
        Handle ``redirect_to`` parameter.
        """
        username = request.POST.get("username")
        password = request.POST.get("password")
        user = authenticate(username=username, password=password)
        if user is None:
            logout(request)
            context = self.get_context_data()
            context["message"] = "Login Failed"
            context["redirect_to"] = request.POST.get("redirect")
            return render(request, self.template_name, context)

        auth_manager = AuthenticationBackendManager()
        auth_manager.fetch_user(user)
        auth_manager.login(user, request)
        return redirect(request.POST.get("redirect", reverse("core:project_list")))


class RegisterView(TemplateView):
    """Handle registration of new heliport users when using the
    :class:`heliport.core.user_logic.user_information_managers.managers.BasicUserInformationManager`."""

    #: HTML template rendered by this View
    template_name = "core/user/register.html"

    def get(self, request, *args, **kwargs):
        """Called by django on http get. Redirect to log in if not using "basic" auth backend setting."""
        if settings.AUTHENTICATION_BACKEND_USED != "basic":
            return redirect("core:login")
        return super().get(request, *args, **kwargs)

    def post(self, request):
        """Create User from provided data. Called by django on http post."""
        if settings.AUTHENTICATION_BACKEND_USED != "basic":
            return redirect("core:login")

        username = request.POST.get("username")
        password = request.POST.get("password")
        password2 = request.POST.get("password2")
        surname = request.POST.get("surname")
        given_name = request.POST.get("given_name")
        email = request.POST.get("email")
        orcid = request.POST.get("orcid")
        message = ""

        if password != password2:
            message = "Passwords don't match"
        if User.objects.filter(username=username).first() is not None:
            message = "Username already exists"
        if message:
            return render(request, self.template_name, context={"message": message})

        user = User.objects.create_user(username, email, password)
        user.first_name = given_name
        user.last_name = surname
        user.save()

        if orcid:
            heliport_user = HeliportUser()
            heliport_user.display_name = (
                AuthenticationBackendManager().get_display_name(user)
            )
            heliport_user.auth_user = user
            heliport_user.set_orcid(orcid)
            heliport_user.save()

        return redirect("core:login")


class LogoutView(View):
    """Handle logging out of HELIPORT."""

    def get(self, request):
        """Log out current user. Called by django on http get."""
        logout(request)
        return redirect("core:login")


class IndexView(View):
    """Handle what happens at HELIPORT root url."""

    def get(self, request):
        """Redirect to :class:`ProjectListView` if user is logged in or :class:`LoginView` otherwise.
        Called by django on http get."""
        if request.user.is_authenticated:
            return redirect("core:project_list")
        return redirect("core:login")


class PublicLandingPageView(DetailView):
    """Handle rendering public landing pages for Digital Objects.
    Normal version is :class:`LandingPageView`. This view exists in addition to that view in order to remove
    functionality to reduce probability of showing non-public information.
    Used by `landing page server <https://codebase.helmholtz.cloud/fwcc/data-management/landing-page-server>`_ to make
    landing pages available outside the internal network.
    """

    #: HTML template rendered by this View
    template_name = "core/digital_object/public_landingpage.html"

    #: Django model handled in this view - used by django ``DetailView``
    model = DigitalObject

    def get_context_data(self, **kwargs):
        """Get relevant infos in addition to the Digital Object itself.
        Called by django to get context when rendering html template."""

        context = super().get_context_data(**kwargs)
        obj = self.object.casted_or_self()
        context["renderers"] = get_unique_renderers(obj)
        context["static_html"] = True
        context["attributes_by_category"] = get_attributes_by_category(
            obj, user=None
        ).items()
        context["host"] = self.request.get_host()
        context["protocol"] = self.request.scheme
        context["favicon_url"] = settings.HELIPORT_CORE_PUBLIC_FAVICON_URL
        no_redirect_param = self.request.GET.get("public_version")
        if no_redirect_param is not None and str(no_redirect_param).lower() != "false":
            context["no_redirect"] = True
        return context

    def get(self, request, *args, **kwargs):
        """Select renderer based on ``ACCEPT`` header and render response. Called by django on http get."""

        digital_object = self.get_object().casted_or_self()

        renderer, error = select_renderer(
            request.GET.get("format"), request.META.get("HTTP_ACCEPT")
        )
        if error is not None:
            return HttpResponse(error, status=406, content_type="text/plain")

        if renderer == LandingPageRenderer or renderer == HtmlRenderer:
            return super().get(request, *args, **kwargs)

        response = renderer.render(digital_object, user=None)
        return response


class LandingPageView(DetailView):
    """Handle rendering landing pages in different formats (including html, rdf, datacite) for Digital Objects.
    For only public information :class:`PublicLandingPageView` also exists."""

    #: HTML template rendered by this View
    template_name = "core/digital_object/landingpage.html"
    #: Django model handled in this view - used by django ``DetailView``
    model = DigitalObject

    def get_context_data(self, **kwargs):
        """Get relevant infos in addition to the Digital Object itself.
        Called by django to get context when rendering html template."""
        context = super().get_context_data(**kwargs)
        context["renderers"] = get_unique_renderers()
        obj = self.object.casted_or_self()
        user = None
        if hasattr(self.request.user, "heliportuser"):
            user = self.request.user.heliportuser
        context["attributes_by_category"] = get_attributes_by_category(
            obj, user
        ).items()
        context["has_write_permission"] = has_write_permission(user, obj)
        context["host"] = self.request.get_host()
        context["protocol"] = self.request.scheme
        return context

    @staticmethod
    def get_custom_url(obj, request):
        """Get the url of a special page in heliport for showing ``obj``. This information is defined on ``obj``
        via a ``get_custom_url()`` function. For example a project is shown as the heliport project graph. (see
        :meth:`heliport.core.models.Project.get_custom_view_url`)"""
        user = None
        if hasattr(request.user, "heliportuser"):
            user = request.user.heliportuser

        if hasattr(obj, "get_custom_view_url"):
            return obj.get_custom_view_url(user)
        else:
            return None

    def get(self, request, *args, **kwargs):
        """
        Renderer selection based on `Accept header <https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Accept>`_
        and handling of special renderers like showing individual page with :meth:`get_custom_url`.

        Called by django on http get.
        """
        digital_object = self.get_object().casted_or_self()
        user = None
        if hasattr(request.user, "heliportuser"):
            user = request.user.heliportuser

        renderer, error = select_renderer(
            request.GET.get("format"), request.META.get("HTTP_ACCEPT")
        )
        if error is not None:
            return HttpResponse(error, status=406, content_type="text/plain")

        if renderer == LandingPageRenderer:
            return super().get(request, *args, **kwargs)
        if renderer == HtmlRenderer:
            custom_url = self.get_custom_url(digital_object, request)
            if custom_url is None:
                return super().get(request, *args, **kwargs)
            else:
                return redirect(custom_url)

        response = renderer.render(digital_object, user)
        return response

    def post(self, request, *args, **kwargs):
        """Handle editing what attributes are publicly shown. Called by django on http post."""
        public_selection = request.POST.getlist("public_attributes")
        obj = self.get_object().casted_or_self()
        user = None
        if hasattr(request.user, "heliportuser"):
            user = request.user.heliportuser
        if not has_write_permission(user, obj):
            return HttpResponse(status=http.client.FORBIDDEN)
        attributes = {
            attr.get_id(obj): attr.bind_to(obj) for attr in get_all_attributes(obj)
        }

        todo = []
        for attr_is_public_select in public_selection:
            is_public = request.POST.get(attr_is_public_select) == "public"
            attr_id = remove_prefix(attr_is_public_select, "public_attributes_")
            attr = attributes.get(attr_id)
            if attr is not None and attr.is_public != is_public:
                # Not setting is_public directly and checking if something changed
                # helps when two attributes share an is_public boolean
                todo.append((attr, is_public))

        for attr, is_public in todo:
            attr.is_public = is_public

        return redirect("core:landing_page", **kwargs)


class LandingPageByPersistentIDView(RedirectView):
    """
    Handle redirecting to landing page when given query parameters ``persistent_id``, ``handle`` or ``doi``.
    """

    #: Keep query string after redirect
    query_string = True

    def get_redirect_url(self, *args, **kwargs):
        """Find the digital object and return url of its landing page. Called by ``RedirectView``."""
        persistent_id = self.request.GET.get("persistent_id")
        handle = self.request.GET.get("handle")
        doi = self.request.GET.get("doi")
        if handle:
            persistent_id = f"https://hdl.handle.net/{handle}"
        if doi:
            persistent_id = f"https://doi.org/{doi}"
        answer_json = self.request.GET.get("json", "false")
        if not persistent_id:
            raise Http404('"persistent_id", "handle" or "doi" parameter not specified')

        digital_object = DigitalObject.objects.filter(
            persistent_id__iexact=persistent_id
        ).first()
        if digital_object is None:
            digital_object = DigitalObject.objects.filter(
                generated_persistent_id__iexact=persistent_id
            ).first()
        if digital_object is None:
            raise Http404("persistent id not found")

        url = reverse("core:landing_page", kwargs={"pk": digital_object.pk})
        if answer_json.lower() != "false":
            url = f"{ url }?json=true"
        return url


class ProjectListView(HeliportLoginRequiredMixin, ListView):
    """Handle listing a users HELIPORT project."""

    #: Django model handled in this view - used by django ``ListView``
    model = Project
    #: HTML template rendered by this View
    template_name = "core/project/project_list.html"

    def get_context_data(self, **kwargs):
        """Called by django to get context when rendering template."""
        context = super().get_context_data(**kwargs)
        context["project_create_urls"] = get_project_create_urls()
        return context

    def get_queryset(self):
        """Called by ListView to get the list to show.
        Query: projects of user that are not deleted; ordered by creation date"""
        user = self.request.user.heliportuser
        return projects_of(user).filter(deleted__isnull=True).order_by("-created")


class ProjectGraphView(HeliportProjectMixin, DetailView):
    """Handle showing the HELIPORT project graph"""

    #: HTML template rendered by this View
    template_name = "core/project/project_graph.html"
    #: Django model handled in this view - used by django ``DetailView``
    model = Project
    #: Expect url parameter "project" instead of "pk"
    pk_url_kwarg = "project"

    def get_context_data(self, **kwargs):
        """Generate heliport project graph.
        As a side effect this sets "activeHeliportProject" to ``project.pk`` which is used in navbar to show current project.
        Called by django to get context when rendering template.
        """
        context = super().get_context_data(**kwargs)
        project = context["project"]
        self.request.session["activeHeliportProject"] = project.pk

        graph = generate_graph(self.object)
        context["graph"] = graph
        sections_count = len(list(graph))
        context["graph_size_large"] = 480 * sections_count
        context["graph_size_small"] = 370 * sections_count

        return context


class ProjectTimelineView(HeliportProjectMixin, TemplateView):
    """Handle showing timeline of all digital objects in a project."""

    #: HTML template rendered by this View
    template_name = "core/project/timeline.html"

    def get_context_data(self, **kwargs):
        """Called by django to get context when rendering template.
        Query all not deleted digital objects in project ordered by creation date.
        Get or create :class:`heliport.core.models.TimelineConfiguration`.
        """
        context = super().get_context_data(**kwargs)
        project = context["project"]
        context["digital_objects"] = project.digitalobject_set.filter(
            deleted__isnull=True
        ).order_by("-created")
        (
            context["timeline_configuration"],
            is_new,
        ) = TimelineConfiguration.objects.get_or_create(project=project)
        return context

    def post(self, request, *args, **kwargs):
        """Handle saving if timeline should be shown in project graph using
        :class:`heliport.core.models.TimelineConfiguration`.
        Called by django on http post.
        """
        project = get_object_or_404(Project, pk=self.kwargs.get("project"))
        timeline_config, is_new = TimelineConfiguration.objects.get_or_create(
            project=project
        )
        timeline_config.show_timeline = request.POST["show_timeline"] == "true"
        timeline_config.save()
        logger.debug(f"show timeline {timeline_config.show_timeline}")
        return JsonResponse({"show_timeline": timeline_config.show_timeline})


class CreateEmptyProjectView(HeliportLoginRequiredMixin, View):
    """Handle creating new HELIPORT project.
    There are no questions asked in this view - The project is created immediately, configuration can be changed later.
    """

    def get(self, request):
        """Create empty project in HELIPORT database. Called by django on http get."""
        project = create_empty_project(request)
        return redirect("core:project_graph", project=project.pk)


class CreateProjectView(HeliportLoginRequiredMixin, TemplateView):
    """Handle creating new HELIPORT project.
    Including options for specifying initial project settings and importing contents of other projects.
    """

    #: HTML template rendered by this View (page for entering initial project settings)
    template_name = "core/project/create_project.html"

    def get_context_data(self, **kwargs):
        """Called by django to get context when rendering template.
        Context includes department information and list of projects of current user."""
        context = super().get_context_data(**kwargs)

        with at_least_one_department():
            context["departments"] = Group.objects.order_by("display_name")
        context["default_department"] = AuthenticationBackendManager().get_group(
            self.request.user
        )
        context["projects"] = projects_of(self.request.user.heliportuser).order_by(
            Lower("label")
        )

        return context

    def post(self, request, *args, **kwargs):
        """
        Create project and set settings.
        If template project is selected, to import digital objects from, redirect to :class:`ImportFromProjectView`.
        Called by django on http post.
        """
        project = create_empty_project(request)
        project.label = request.POST.get("label")
        project.description = request.POST.get("description")
        project.group = get_object_or_404(Group, pk=request.POST.get("department"))
        project.save()
        template_project = request.POST.get("template")
        if template_project:
            url = reverse("core:import_project", kwargs=dict(project=project.pk))
            return redirect(f"{url}?from={template_project}")
        else:
            return redirect("core:project_graph", project=project.pk)


class UserSettingsView(HeliportLoginRequiredMixin, TemplateView):
    """Handle showing and editing settings of a HELIPORT user. (most information is read only)"""

    #: HTML template rendered by this View
    template_name = "core/user/user_settings.html"

    def get_context_data(self, **kwargs):
        """Called by django to get context when rendering template.
        Context includes the user and information for HELIPORT API token creation form (like the current time)
        """
        context = super().get_context_data(**kwargs)
        user = self.request.user

        context["heliport_user"] = user.heliportuser
        context["now"] = timezone.now()
        context["tomorrow"] = (timezone.now() + timedelta(days=1)).strftime("%Y-%m-%d")
        context["tokens"] = knox_models.AuthToken.objects.filter(user=user)

        return context

    def post(self, request, *args, **kwargs):
        """Set orcid for user. Called by django on http post."""
        user = request.user.heliportuser
        orcid = request.POST.get("orcid")
        user.set_orcid(orcid)
        user.save()

        return redirect("core:user_settings")


class RemoteServerLoginView(HeliportLoginRequiredMixin, DetailView):
    """Show username and password prompt for an SSH login and put public SSH key in allowed hosts on the server
    specified in the SSH login (:class:`heliport.core.models.LoginInfo`) while private key is stored encrypted in the
    HELIPORT database. When the key of an SSH LoginInfo is on the remote server we say it is "connected".

    The corresponding view for logout is the :class:`RemoteServerLogoutView`.
    """

    #: Django model handled in this view - used by django ``DetailView``
    model = LoginInfo
    #: HTML template rendered by this View
    template_name = "core/user/remote_server_login.html"

    @staticmethod
    def validate_login(login, request):
        """Check if SSH :class:`heliport.core.models.LoginInfo` can be connected.
        This is the case if the login info belongs to the user and all necessary SSH indirections are connected.
        """

        if request.user.heliportuser != login.user:
            return redirect(reverse("core:login"))

        logger.debug("check if all parents are connected")
        try:
            connect_first = login.get_unconnected_parents()
        except ValueError:
            return HttpResponse("<b> ERROR: </b> infinite chain of ssh connections")
        if connect_first:
            request.session[
                "message"
            ] = f"{login.name} can not be connected. Please connect {connect_first[0]} first!"
            return redirect("core:remote_server_login", pk=connect_first[0].pk)

        return None

    def get_context_data(self, **kwargs):
        """Called by django to get context when rendering template. In addition to the LoginInfo
        (automatically included by ``DetailView``) include error message in context."""
        context = super().get_context_data(**kwargs)
        context["message"] = self.request.session.get("message", "")
        self.request.session["message"] = ""
        return context

    def post(self, request, *args, **kwargs):
        """Connect LoginInfo. Called by django on http post.

        As a side effect the username is stored in the ``LoginInfo``. This removes the need to enter the username
        twice (when creating the login info and when connecting the login info). The username is used e.g. for searching
        login infos.
        """
        logger.debug("user requested ssh key connection")
        login_info = self.get_object()
        error = self.validate_login(login_info, request)
        if error is not None:
            return error

        username = request.POST.get("username")
        password = request.POST.get("password")
        login_info.username = username
        login_info.save()

        error = deploy_ssh_key(login_info, username, password)
        if error is not None:
            messages.error(request, error)
        return redirect(reverse("core:user_login_list"))


class RemoteServerLogoutView(HeliportLoginRequiredMixin, View):
    """
    Handle removing SSH key from external server. For more details see :class:`RemoteServerLoginView`.
    This view does not render a template. The key is deleted without any questions asked and the user is redirected to
    :class:`LoginInfoView`.
    """

    def get(self, request, *args, **kwargs):
        """Disconnect SSH :class:`heliport.core.models.LoginInfo`. Called by django on http get."""
        logger.debug("user requested ssh disconnect")
        login_info = get_object_or_404(LoginInfo, pk=kwargs.get("pk"))
        error = RemoteServerLoginView.validate_login(login_info, request)
        if error is not None:
            return error

        delete_ssh_key(login_info)
        return redirect(reverse("core:user_login_list"))


class LoginInfoView(HeliportLoginRequiredMixin, CreateView):
    """
    Handle creating/showing login details of other web services in HELIPORT so that HELIPORT can make requests on
    behalf of the user. There exists a very similar view for editing/showing such login details:
    :class:`LoginInfoUpdateView`.
    """

    #: Django model handled in this view - used by django ``CreateView``
    model = LoginInfo
    #: HTML template rendered by this View
    template_name = "core/user/login_infos.html"
    #: Used by django ``CreateView`` to generate input form for creating new login infos
    fields = ["name", "type", "username", "key", "machine", "via", "slurm"]

    def get_success_url(self):
        """Called by django to get url to redirect to after successful :class:`heliport.core.models.LoginInfo` creation

        :return: URL of this view itself.
        """
        return reverse("core:user_login_list")

    def get_context_data(self, **kwargs):
        """Called by django to get context when rendering template."""
        user = self.request.user.heliportuser
        context = super().get_context_data(**kwargs)
        context["logins"] = user.logininfo_set.all()
        context["login_type"] = self.request.GET.get("login_type")
        context["login_types"] = LoginInfo.LoginTypes
        context["ssh_logins"] = user.logininfo_set.filter(type=LoginInfo.LoginTypes.SSH)
        context["update"] = False
        return context

    def post(self, request, *args, **kwargs):
        """Handle deletion of login infos here and creation in superclass ``CreateView``.
        Called by django on http post."""
        login_to_remove = request.POST.get("remove")
        if login_to_remove:
            login_info = get_object_or_404(LoginInfo, pk=login_to_remove)
            if login_info.user != self.request.user.heliportuser:
                return redirect(reverse("core:login"))
            login_info.delete()
            return redirect(self.get_success_url())
        else:
            return super().post(request, *args, **kwargs)

    def form_valid(self, form):
        """Called by django ``CreateView`` after form has been deemed valid. Set additional attributes before saving LoginInfo:

        - **user** - current heliport user added automatically (not from input form)
        - **password** - password field can not be generated automatically in form because `password` is a python property and not a django model field
        """
        form.instance.user = self.request.user.heliportuser
        form.instance.password = self.request.POST.get("password", "")
        return super().form_valid(form)


class LoginInfoUpdateView(HeliportLoginRequiredMixin, UpdateView):
    """
    Handle editing/showing login details, of other web services, in HELIPORT so that HELIPORT can make requests on
    behalf of the user. There exists a very similar view for creation/showing such login details:
    :class:`LoginInfoView`.
    """

    #: Django model handled in this view - used by django ``UpdateView``
    model = LoginInfo
    #: HTML template rendered by this View
    template_name = "core/user/login_infos.html"
    #: Used by django ``UpdateView`` to generate input form for editing login infos
    fields = ["name", "type", "username", "key", "machine", "via", "slurm"]

    def get_success_url(self):
        """Redirect to :class:`LoginInfoView` after updating login info"""
        return reverse("core:user_login_list")

    def get_context_data(self, **kwargs):
        """Called by django to get context when rendering template."""
        user = self.request.user.heliportuser
        context = super().get_context_data(**kwargs)
        context["logins"] = user.logininfo_set.all()
        context["login_type"] = self.request.GET.get("login_type", self.object.type)
        context["login_types"] = LoginInfo.LoginTypes
        context["ssh_logins"] = user.logininfo_set.filter(type=LoginInfo.LoginTypes.SSH)
        context["update"] = True
        return context

    def form_valid(self, form):
        """Assert that :class:`heliport.core.models.LoginInfo` belongs to current user and set password.
        See :meth:`LoginInfoView.form_valid` for more details."""
        if form.instance.user != self.request.user.heliportuser:
            return redirect(reverse("core:login"))
        form.instance.password = self.request.POST.get("password", "")
        return super().form_valid(form)


class ProjectView(HeliportProjectMixin, DetailView):
    """
    Shows general information about a HELIPORT project ("Project Configuration").
    This includes key value pairs like the Title, description or handle of a project
    and information about project participants.

    Basic project actions, like deleting the project, can be done via this view too.
    """

    #: Django model handled in this view - used by django ``DetailView``
    model = Project
    #: HTML template rendered by this View
    template_name = "core/project/project_update.html"
    #: In :mod:`heliport.core.urls` project id is called "project" not "pk"
    pk_url_kwarg = "project"

    def get_context_data(self, **kwargs):
        """
        Called by django to get context when rendering template.

        The context includes:

            - "isOwner", "update" to configure the template for what should be shown
            - preprocessed information about project participants
            - "groups" the list of all departments to display a select box
        """

        context = super().get_context_data(**kwargs)
        context["isOwner"] = self.object.owner == self.request.user.heliportuser
        with at_least_one_department():
            context["groups"] = Group.objects.order_by("display_name")
        self.request.session["activeHeliportProject"] = self.object.pk
        context["co_owners"] = sorted(context["project"].co_owners.all(), key=str)
        context["contributions"] = sorted(
            context["project"].contribution_set.all(), key=str
        )
        context["contribution_types"] = Contribution.ContributionTypes.choices
        context["contribution_type_default"] = Contribution._meta.get_field(
            "type"
        ).get_default()
        context["update"] = False
        return context

    def post(self, request, *args, **kwargs):
        """
        Handles one of:

            - Delete project
            - Add file system user group to project
            - Add / Remove co-owner
            - Add all members from a department as co-owners
            - Add / Delete Contribution
            - Edit contribution type (request is send to this view but ui is rendered in :class:`ProjectUpdateView`)

        Called by django on http post.
        """

        remove = request.POST.get("remove")
        user_id = request.POST.get("add")
        add_group = request.POST.get("add_group")
        delete_owner = self.request.POST.get("user")
        group = self.request.POST.get("group_to_import")
        contribution_delete = request.POST.get("contribution_delete")
        contributor_name = request.POST.get("contributor_name")
        contributor_affiliation = request.POST.get("contributor_affiliation")
        contributor_orcid = request.POST.get("contributor_orcid")
        contribution_type = request.POST.get("contribution_type")
        contribution_modify = request.POST.get("contribution_modify")
        contribution_type_update = request.POST.get("contribution_type_update")

        project = self.get_object()
        heliport_group = project.heliport_groups.first()

        if remove:
            if project.owner == request.user.heliportuser:
                project.deleted = timezone.now()
                project.save()
            return redirect(reverse("core:project_list"))

        elif add_group:
            if not project.heliport_groups.all():
                new_group = HeliportGroup()
                new_group.save()
                new_group.group_name = (
                    f"{settings.HELIPORT_CORE_GROUP_NAME_PREFIX}-{new_group.pk}"
                )
                new_group.save()
                logger.info(f"CREATED GROUP {new_group}")
                project.heliport_groups.add(new_group)
                for user in project.members():
                    logger.info(f"ADD MEMBER {user} TO GROUP {new_group}")

        elif user_id:
            user = get_object_or_404(HeliportUser, pk=int(user_id))
            if project.owner != user:
                if heliport_group is not None:
                    logger.info(f"ADD MEMBER {user} TO GROUP {heliport_group}")
                project.co_owners.add(user)

        elif delete_owner:
            user = get_object_or_404(HeliportUser, pk=int(delete_owner))
            if heliport_group is not None:
                logger.info(f"REMOVE MEMBER {user} FROM GROUP {heliport_group}")
            project.co_owners.remove(user)

        elif group:
            db_group = get_object_or_404(Group, pk=int(group))
            for db_user in AuthenticationBackendManager().get_users_by_group(db_group):
                if db_user != project.owner:
                    project.co_owners.add(db_user)

        elif contribution_delete:
            # This will leave HeliportUser objects that are essentially useless. But at
            # this point we can't really decide if there was another purpose for this
            # user.
            contribution = get_object_or_404(Contribution, pk=int(contribution_delete))
            contribution.delete()

        elif contributor_name:
            user = HeliportUser(
                display_name=contributor_name,
                affiliation=contributor_affiliation,
            )
            if contributor_orcid:
                user.set_orcid(contributor_orcid)
            user.save()
            Contribution(
                contributor=user,
                contribution_to=project,
                type=contribution_type,
            ).save()

        elif contribution_modify and contribution_type_update:
            contribution = get_object_or_404(Contribution, pk=int(contribution_modify))
            contribution.type = contribution_type_update
            contribution.save()

        return redirect("core:project", project=project.pk)


class ProjectUpdateView(HeliportProjectMixin, UpdateView):
    """
    Handles updating project title, description and department.
    This is rendering the same template as :class:`ProjectView`
    """

    #: Django model handled in this view - used by django ``UpdateView``
    model = Project
    #: HTML template rendered by this View
    template_name = "core/project/project_update.html"
    #: Used by django ``UpdateView`` to generate input form for editing project
    fields = ["label", "description", "group"]
    #: In :mod:`heliport.core.urls` project id is called "project" not "pk"
    pk_url_kwarg = "project"

    def get_context_data(self, **kwargs):
        """See :meth:`ProjectView.get_context_data` for more information."""
        context = super().get_context_data(**kwargs)
        context["isOwner"] = self.object.owner == self.request.user.heliportuser
        with at_least_one_department():
            context["groups"] = Group.objects.order_by("display_name")
        self.request.session["activeHeliportProject"] = self.object.pk
        context["co_owners"] = sorted(context["project"].co_owners.all(), key=str)
        context["contributions"] = sorted(
            context["project"].contribution_set.all(), key=str
        )
        context["contribution_types"] = Contribution.ContributionTypes.choices
        context["update"] = True
        return context

    def get_success_url(self):
        """Return to :class:`ProjectView` after updating and saving project."""
        return reverse("core:project", kwargs={"project": self.object.pk})


class DigitalObjectAutocompleteView(HeliportLoginRequiredMixin, View):
    """
    Handles getting information for autocomplete inputs for :class:`heliport.core.models.DigitalObject`
    """

    def get(self, request, *args, **kwargs):
        """
        Called by django on http get. Returns json list like::

            [
                {"label": "aa", "pk": 1},
                {"label": "xa", "pk": 55}
            ]

        Excludes deleted objects and respects user rights.
        Maximum dumber of results is configurable in settings.py ``HELIPORT_CORE_AUTOCOMPLETE_COUNT``

        Supports the following get parameters:

            - ``type`` - Restrict autocomplete to a type. See :func:`heliport.core.utils.queries.digital_objects_by_type` for allowed values. ``...?type=1&type=2`` results in union of results to ``...?type=1`` and ``...?type=2``.
            - ``q`` or ``term`` - search term for label and persistent_id
            - ``exact`` - false (default): case-insensitive and allow part match; case_insensitive: case-insensitive full match; true: case-sensitive full match. psersistent_id is always matched exact
        """
        query = ListOfQuerySets()

        obj_types = request.GET.getlist("type")
        if obj_types:
            for obj_type in obj_types:
                query.extend(digital_objects_by_type(obj_type))
        else:
            query.all_digital_objects()

        auth_user = request.user
        user = None
        if hasattr(auth_user, "heliportuser"):
            user = auth_user.heliportuser
        if user is None:
            query.filter(owner__isnull=True)
        else:
            query.filter(Q(owner__isnull=True) | Q(owner=user) | Q(co_owners=user))

        query.filter(deleted__isnull=True)

        pattern = request.GET.get("q") or request.GET.get("term", "")
        if pattern:
            exact = str(request.GET.get("exact", "false")).lower()
            if exact == "false":
                q = Q(label__icontains=pattern)
            elif exact == "case_insensitive":
                q = Q(label__iexact=pattern)
            else:
                q = Q(label__exact=pattern)

            query.filter(q | Q(persistent_id=pattern))

        return JsonResponse(
            [
                {"label": obj.label, "pk": obj.digital_object_id}
                for obj in query.as_list(settings.HELIPORT_CORE_AUTOCOMPLETE_COUNT)
            ],
            safe=False,
        )


def get_users_autocomplete(request, **kwargs):
    """
    View that handles user autocomplete. The query string is taken from get parameter `q`.
    Respects ``HELIPORT_CORE_AUTOCOMPLETE_COUNT`` in settings.py.

    Returns json like::

        [
            {"name": "ax", "pk": 1},
            {"name": "xyyy", "pk": 55}
        ]
    """

    query = request.GET.get("q")

    if query:
        sql_query = HeliportUser.objects.filter(display_name__icontains=query)
    else:
        sql_query = HeliportUser.objects.all()
    sql_query = sql_query.order_by(F("auth_user__last_login").desc(nulls_last=True))

    results = []
    names = set()
    for guy in sql_query[: settings.HELIPORT_CORE_AUTOCOMPLETE_COUNT]:
        if guy.display_name not in names:
            results.append({"name": guy.display_name, "pk": guy.pk})
            names.add(guy.display_name)

    return JsonResponse(results, safe=False)


def get_project_list(request):
    """
    View that handles getting projects of a user as json. This is used in the HELIPORT navbar.

    Responds with json like::

        [
            {"name": "project label", "pk": project pk, "url": "url to project graph"}
        ]
    """

    if request.user.is_authenticated:
        user = request.user.heliportuser
        projects = list(projects_of(user))
        data = [
            {
                "name": project.label,
                "pk": project.pk,
                "url": reverse("core:project_graph", kwargs={"project": project.pk}),
            }
            for project in projects
            if project.deleted is None
        ]
        return JsonResponse(data, safe=False)
    return JsonResponse([], safe=False)


class FaviconView(View):
    """Returns the HELIPORT logo as icon. To be shown in browser tab."""

    def get(self, request, **kwargs):
        """Read icon and return it as HttpResponse."""
        icon = finders.find("core/img/favicon.ico")
        with open(icon, "rb") as f:
            response = HttpResponse(f.read(), content_type="image/ico")
        return response


class DigitalObjectQRView(DetailView):
    """Generate QR code for digital object specified in url"""

    model = DigitalObject

    def get(self, request, **kwargs):
        """Called by django on http get."""
        obj = self.get_object()
        qr = qrcode.make(obj.identifier_link)
        with io.BytesIO() as byte_array:
            qr.save(byte_array, format="PNG")
            response = HttpResponse(byte_array.getvalue(), content_type="image/png")
        return response


class ImageView(HeliportObjectMixin, DetailView):
    """View for showing :class:`heliport.core.models.Image` s to use them in html."""

    model = Image

    def get(self, request, *args, **kwargs):
        """
        Image pk is taken from url. This view may be cached (configured in :mod:`heliport.core.urls`).
        Called by django on http get.
        """
        image = self.get_object()
        byte_data = image.download_data()

        if image.status == image.ImageStates.UPLOADING:
            # AssertionError here causes data not to be cached
            assert byte_data != b"", "image not uploaded"
        return HttpResponse(byte_data, content_type=image.mimetype)


class TakePictureView(HeliportLoginRequiredMixin, TemplateView):
    """Handles taking pictures in HELIPORT these pictures are stored as :class:`heliport.core.models.Image`"""

    #: HTML template rendered by this View
    template_name = "core/digital_object/take_picture.html"

    def get_context_data(self, **kwargs):
        """
        Generate default values for storing image and include them in context.
        Support ``next`` get parameter to redirect to arbitrary page after image has been taken.
        """

        context = super().get_context_data(**kwargs)
        context["project"] = get_object_or_404(Project, pk=self.kwargs.get("project"))
        last_image = (
            Image.objects.filter(login__user__auth_user=self.request.user)
            .order_by("-created")
            .first()
        )
        context["default_login"] = None if last_image is None else last_image.login
        context["default_directory"] = (
            "" if last_image is None else last_image.directory
        )
        context["logins"] = LoginInfo.objects.filter(
            user__auth_user=self.request.user, type=LoginInfo.LoginTypes.SSH
        )
        context["next"] = self.request.GET.get("next")
        return context

    def post(self, request, *args, **kwargs):
        """
        Create new :class:`heliport.core.models.Image` with user supplied metadata
        and upload image itself to file system specified by login info (selected by user).

        Return redirect to image list or view specified by "next" parameter.
        If "next" parameter ends with "=" append pk of newly taken image.
        For example ``https://some/url?image=`` becomes ``https://some/url?imgae=13`` if ``13`` is the newly taken image.

        Called by django on http post.
        """
        project = get_object_or_404(Project, pk=self.kwargs.get("project"))
        login = get_object_or_404(LoginInfo, pk=request.POST.get("login"))
        image_data = request.POST["data_url"]
        assert image_data.startswith("data:image/"), "not an image"

        new_image = Image()
        new_image.filename = request.POST["filename"]
        new_image.directory = request.POST["directory"]
        new_image.login = login
        new_image.label = request.POST.get("label", "")
        new_image.description = request.POST.get("description", "")
        new_image.category_str = settings.IMAGE_NAMESPACE
        new_image.save()
        new_image.register_handle_if_applicable(request)
        new_image.projects.add(project)
        new_image.upload_image(image_data)

        redirect_url = request.GET.get("next", "")
        if redirect_url.endswith("="):
            redirect_url = f"{redirect_url}{new_image.pk}"
        if not redirect_url:
            redirect_url = reverse("core:image_list", kwargs={"project": project.pk})

        return redirect(redirect_url)


class ImagesView(HeliportProjectMixin, TemplateView):
    """List all :class:`heliport.core.models.Image` s in a project."""

    #: HTML template rendered by this View
    template_name = "core/digital_object/picture_list.html"

    def get_context_data(self, **kwargs):
        """
        Called by django to get context when rendering template.
        Includes "images", the list of relevant images, in context
        """
        user = get_object_or_404(HeliportUser, auth_user=self.request.user)
        project = get_object_or_404(Project, pk=self.kwargs["project"])
        context = super().get_context_data(**kwargs)
        context["images"] = Image.objects.filter(
            is_object_member(user), deleted__isnull=True, projects=project
        ).distinct()
        return context

    def post(self, request, *args, **kwargs):
        """Handles image deletion. Called by django on http post."""
        image_to_delete = get_object_or_404(Image, pk=request.POST.get("delete"))
        image_to_delete.deleted = timezone.now()
        image_to_delete.save()
        return redirect("core:image_list", project=self.kwargs["project"])


class TagsView(HeliportObjectListView):
    #: Specify that this View shows lists of class :class:`heliport.core.models.Tag`.
    model = Tag
    category = settings.TAG_NAMESPACE
    register_handle = True
    columns = [
        ("ID", "tag_id", "small"),
        ("Label", "label", "small"),
        ("Property", "attribute", "normal"),
        ("Value", "value", "normal"),
        ("Used on", "requires_type_display", "small"),
    ]
    actions = [
        ("Edit", "action_edit", "link_secondary"),
        ("Delete", "action_delete", "danger"),
    ]
    list_url = "core:tag_list"
    update_url = "core:tag_update"
    list_heading = "Tags"
    create_heading = "Add a Tag"
    add_text = [
        "A tag describes that something has <Property> set to <Value>. "
        'For example a tag "Raw Data" describes "type" is set to "Raw Data". '
        'This "Raw Data" tag could be constrained to be USED ON "data_source". '
    ]
    edit_tags = True

    def register_form_columns(self, form: FormDescription):
        """Register custom form columns. Called by :meth:`heliport.core.custom_views.HeliportObjectListView.get_table`"""
        vocab = Vocabulary()
        form.add_column("Label", "label", attrs={"required": True})
        form.add_column(
            "Color", "html_color", attrs={"widget": "color", "required": True}
        )
        form.add_column(
            "Property",
            "attribute",
            attrs={"widget": "digital_object", "type": vocab.Property.pk},
        )
        form.add_column("Value", "value", attrs={"widget": "digital_object"})
        form.add_column(
            "Used on",
            "requires_type",
            attrs={"widget": "digital_object", "type": vocab.Class.pk},
        )

    def set_update_attributes(self, obj, form_fields, object_saved=False):
        """Custom implementation of setting attributes."""
        super().set_update_attributes(obj, form_fields)
        vocab = Vocabulary()
        if not hasattr(obj, "attribute") or obj.attribute is None:
            attribute_label = self.request.POST.get("form_data_object_label_attribute")
            if attribute_label:
                attribute = DigitalObject(label=attribute_label)
                attribute.category_str = settings.PROPERTY_NAMESPACE
                attribute.save()
                assert_relation(attribute, vocab.type, vocab.Property)
                attribute.projects.add(self.project)
                attribute.register_handle_if_applicable(self.request)
            else:
                attribute = vocab.has_tag
            obj.attribute = attribute

        if not hasattr(obj, "value") or obj.value is None:
            value_label = self.request.POST.get("form_data_object_label_value")
            if value_label:
                value = DigitalObject(label=value_label)
                if obj.attribute == vocab.type:
                    value.category_str = settings.CLASS_NAMESPACE
                else:
                    value.category_str = settings.DEFAULT_OBJECT_NAMESPACE
                value.save()
                if obj.attribute == vocab.type:
                    assert_relation(value, vocab.type, vocab.Class)
                value.projects.add(self.project)
                value.register_handle_if_applicable(self.request)
            else:
                obj.save()
                value = obj
            obj.value = value


class TagJsonView(HeliportProjectMixin, DetailView):
    """
    Handle Querying and updating tags of a digital object.
    This view is used in javascript of tag_manager.html
    """

    #: Django model handled in this view - used by django ``DetailView``
    model = Project
    #: In :mod:`heliport.core.urls` project id is called "project" not "pk"
    pk_url_kwarg = "project"

    def get(self, request, *args, **kwargs):
        """
        Return list of tags including "label", "tag_id", "color", "foreground", ... for a given DigitalObject
        in the context of the project.

        Called by django on http get.
        """
        project = self.get_object()
        digital_object = get_object_or_404(
            DigitalObject, pk=self.request.GET["pk"], projects=project
        )
        return JsonResponse(self.get_all_tags(digital_object, project), safe=False)

    @staticmethod
    def get_all_tags(digital_object, project):
        """
        Get direct (has_tag property) and indirect (There exists a tag with predicate-object-combination also
        found in attributes of ``digital_object`` in ``project``) Tags. Tags are returned as dict of "label",
        "tag_id", "attribute", "value", "requires_type", "color", "foreground".
        """
        tag_ids = set(
            DigitalObjectRelation.objects.filter(
                subject=digital_object,
                predicate=Vocabulary().has_tag,
                object__deleted__isnull=True,
            ).values_list("object", flat=True)
        )
        tags = {Tag.objects.get(digital_object_id=tag_id) for tag_id in tag_ids}
        for possible_tag in Tag.objects.filter(projects=project, deleted__isnull=True):
            if DigitalObjectRelation.objects.filter(
                subject=digital_object,
                predicate=possible_tag.attribute,
                object=possible_tag.value,
            ).exists():
                tags.add(possible_tag)
        return [
            {
                "label": tag.label,
                "tag_id": tag.tag_id,
                "attribute": tag.attribute_id,
                "value": tag.value_id,
                "requires_type": tag.requires_type_id,
                "color": tag.html_color,
                "foreground": tag.html_foreground_color,
            }
            for tag in sorted(tags, key=lambda t: t.label)
        ]

    def post(self, request, *args, **kwargs):
        """
        Adds (or removes when "delete" is set to a truthy value) a tag to a digital object specified by "pk".
        If "check_type" is set to a truthy value the tag is only set when digital object has the correct type for tag.
        The tag can be explicitly specified by "tag" or created from "label", "attribute_pk", "value_pk".
        "attribute_label" and "value_label" are also allowed and their information is combined with attribute_pk and
        value_pk using :func:`heliport.core.models.digital_object_form_label_and_pk`.

        :return: Same as :meth:`get`

        Called by django on http post.
        """
        project = self.get_object()
        vocab = Vocabulary()
        digital_object = get_object_or_404(
            DigitalObject, pk=self.request.POST["pk"], projects=project
        )
        if "tag" in request.POST or request.POST.get("delete"):
            tag = get_object_or_404(Tag, pk=request.POST.get("tag"))
        else:
            label = request.POST.get("label", "new tag")
            tag = Tag.objects.filter(
                projects=project, label=label, deleted__isnull=True
            ).first()
            if tag is None:
                color = pick_color_not_in(
                    [
                        t.rgb_color
                        for t in Tag.objects.filter(
                            projects=project, deleted__isnull=True
                        )
                    ]
                )
                attribute = digital_object_form_label_and_pk(
                    request.POST["attribute_label"], request.POST["attribute_pk"]
                )
                if (
                    attribute is None and request.POST["attribute_label"]
                ):  # maybe use heliport.core.utils.query?
                    attribute = DigitalObject(label=request.POST["attribute_label"])
                    attribute.category_str = settings.PROPERTY_NAMESPACE
                    attribute.save()
                    attribute.projects.add(project)
                    assert_relation(attribute, vocab.type, vocab.Property)
                    attribute.register_handle_if_applicable(request)
                value = digital_object_form_label_and_pk(
                    request.POST["value_label"], request.POST["value_pk"]
                )
                if (
                    value is None and request.POST["value_label"]
                ):  # maybe use heliport.core.utils.query?
                    value = DigitalObject(label=request.POST["value_label"])
                    if attribute == vocab.type:
                        value.category_str = settings.CLASS_NAMESPACE
                    else:
                        value.category_str = settings.DEFAULT_OBJECT_NAMESPACE
                    value.save()
                    if attribute == vocab.type:
                        assert_relation(value, vocab.type, vocab.Class)
                    value.projects.add(project)
                    value.register_handle_if_applicable(request)

                tag = Tag.objects.create(label=label)
                tag.rgb_color = color
                tag.attribute = attribute
                tag.category_str = settings.TAG_NAMESPACE
                tag.value = value
                if tag.attribute is None:
                    tag.attribute = vocab.has_tag
                tag.save()
                if tag.value is None:
                    tag.value = tag
                tag.save()
                tag.projects.add(project)
                tag.register_handle_if_applicable(request)

        tag.update_usage_date()

        if (
            request.POST.get("check_type")
            and not request.POST.get("delete")
            and tag.requires_type is not None
        ):
            if not DigitalObjectRelation.objects.filter(
                subject=digital_object,
                predicate=vocab.type,
                object=tag.requires_type,
            ).exists():
                return JsonResponse(
                    self.get_all_tags(digital_object, project), safe=False
                )

        attribute = tag.attribute
        value = tag.value
        if attribute is None:
            attribute = vocab.has_tag
        if value is None:
            value = tag
        if not request.POST.get("delete"):
            assert_relation(digital_object, attribute, value)
        else:
            assert_not_relation(digital_object, attribute, value)
            assert_not_relation(digital_object, vocab.has_tag, tag)
        return JsonResponse(self.get_all_tags(digital_object, project), safe=False)


class TagAutocompleteView(HeliportProjectMixin, DetailView):
    """Handle autocompletion when specifying tags."""

    #: Django model handled in this view - used by django ``DetailView``
    model = Project
    #: In :mod:`heliport.core.urls` project id is called "project" not "pk"
    pk_url_kwarg = "project"

    def get(self, request, *args, **kwargs):
        """
        Partial input can be submitted via "q" or "term".
        If "pk" is specified, only tags allowed for this DigitalObject are returned unless there are none.

        Tags in result json are returned with "label" and "pk" ordered by least recently used.

        Called by django on http get.
        """
        project = self.get_object()
        digital_object = DigitalObject.objects.filter(
            projects=project, pk=request.GET.get("pk")
        ).first()
        pattern = request.GET.get("q") or request.GET.get("term", "")

        tags = Tag.objects.filter(
            projects=project,
            deleted__isnull=True,
            label__icontains=pattern,
        )

        if digital_object is not None:
            types = DigitalObjectRelation.objects.filter(
                subject=digital_object, predicate=Vocabulary().type
            ).values_list("object", flat=True)
            filtered_tags = tags.filter(
                Q(requires_type__in=types) | Q(requires_type__isnull=True)
            )
            if len(filtered_tags) > 0:
                tags = filtered_tags

        return JsonResponse(
            [
                {"label": tag.label, "pk": tag.pk}
                for tag in tags.order_by("-last_use_date")
            ],
            safe=False,
        )


class SubprojectsView(HeliportProjectMixin, DetailView):
    """Handle other HELIPORT projects that are part of the current project."""

    #: HTML template rendered by this View
    template_name = "core/project/subprojects.html"
    #: Django model handled in this view - used by django ``DetailView``
    model = Project
    #: In :mod:`heliport.core.urls` project id is called "project" not "pk"
    pk_url_kwarg = "project"

    def get_context_data(self, **kwargs):
        """Called by django to get context when rendering template.
        Subprojects, all projects of user and user are included in context"""
        context = super().get_context_data(**kwargs)
        context["subprojects"] = Project.objects.filter(projects=self.object)
        context["user_projects"] = projects_of(self.request.user.heliportuser)
        context["heliport_user"] = self.request.user.heliportuser
        return context

    def post(self, request, **kwargs):
        """ "add" or "remove" a subproject"""
        project = self.get_object()
        user = self.request.user.heliportuser
        to_remove = request.POST.get("remove")
        to_add = request.POST.get("add")
        if to_remove:
            subproject = get_object_or_404(Project, pk=to_remove)
            assert user in project.members()
            subproject.projects.remove(project)
        if to_add:
            subproject = get_object_or_404(Project, pk=to_add)
            assert user in project.members()
            assert user in subproject.members()
            subproject.projects.add(project)
        return redirect("core:subprojects", project=project.pk)


class ImportFromProjectView(HeliportProjectMixin, TemplateView):
    """Handle importing or copying digital objects from another project"""

    #: HTML template rendered by this View
    template_name = "core/project/import_from.html"

    def test_func(self):
        """
        Called by django to test if user has access to this view.
        In addition to normal test if user has access to current project also test if user has read access to project
        where things should be imported from.
        """
        if not super().test_func():
            return False

        other_project = get_object_or_404(Project, pk=self.request.GET.get("from"))
        user = self.request.user.heliportuser
        return has_read_permission(user, other_project)

    def get_context_data(self, **kwargs):
        """
        Called by django to get context when rendering template.

        Include list of object that can be imported in context, including their label, types and id.
        Also for each object, the "default_type" (completely different meaning than "types";
        one of ignore, copy, import) and "copy_disabled" (bool; not every object is copyable) is included.
        """
        vocab = Vocabulary()

        def get_default_import_type(obj, past_count):
            """compute default type based on usage pattern in the past"""
            copy_count = DigitalObjectRelation.objects.filter(
                object=obj, predicate=vocab.is_version_of
            ).count()
            import_count = obj.projects.count() - 1
            if copy_count * 4 <= past_count and import_count * 4 <= past_count:
                return "ignore"
            elif copy_count >= import_count:
                return "copy"
            else:
                return "import"

        context = super().get_context_data(**kwargs)
        other_project = get_object_or_404(Project, pk=self.request.GET.get("from"))
        past_import_projects = set()
        for obj in DigitalObject.objects.filter(projects=other_project):
            past_import_projects.update(obj.projects.all())
        for relation in DigitalObjectRelation.objects.filter(
            object__projects=other_project, predicate=vocab.is_version_of
        ):
            past_import_projects.update(relation.subject.projects.all())
        past_import_projects.discard(other_project)

        context["objects_to_import"] = sorted(
            [
                {
                    "label": obj.label or obj.description,
                    "types": types(obj.casted_or_self(), bound_attributes=True),
                    "default_type": get_default_import_type(
                        obj, len(past_import_projects)
                    ),
                    "id": obj.digital_object_id,
                    "copy_disabled": hasattr(obj.casted_or_self(), "disable_copy")
                    and obj.casted_or_self().disable_copy,
                }
                for obj in other_project.digitalobject_set.filter(deleted__isnull=True)
                if has_read_permission(self.request.user.heliportuser, obj)
            ],
            key=lambda x: str(x["label"] or "").lower(),
        )
        context["objects_to_import_len"] = len(context["objects_to_import"])
        context["from"] = other_project
        return context

    def post(self, request, *args, **kwargs):
        """Do the importing / copying. "import_objects" is the list of all objects.
        E.g. for obj with pk 7 a value import-type-7 (import, copy, ignore) specifies the kind of import."""
        user = self.request.user.heliportuser
        current_project = get_object_or_404(Project, pk=self.kwargs["project"])
        if not has_write_permission(user, current_project):
            return redirect("core:login")

        for obj_id in self.request.POST.getlist("import_objects"):
            obj = get_object_or_404(DigitalObject, digital_object_id=obj_id)
            obj = obj.casted_or_self()
            if not has_read_permission(user, obj):
                continue
            import_type = self.request.POST.get(f"import-type-{ obj_id }")
            if import_type == "import":
                obj.projects.add(current_project)
            elif import_type == "copy" and not (
                hasattr(obj, "copy_disabled") and obj.copy_disabled
            ):
                new_obj = obj.create_copy()
                if new_obj.category:
                    new_obj.register_handle_if_applicable(self.request)
                new_obj.projects.add(current_project)

        return redirect("core:project_graph", project=current_project.pk)


class SearchBaseView(HeliportLoginRequiredMixin, TemplateView):
    """
    Base view for the HELIPORT search feature.
    This view uses javascript to get the actual search results from each apps search view.
    """

    #: HTML template rendered by this View
    template_name = "core/search/search_base.html"

    def get_context_data(self, **kwargs):
        """Called by django to get context when rendering template. Includes urls of all search views."""
        context = super().get_context_data(**kwargs)
        context["search_urls"] = list(get_search_urls())
        context["query"] = self.request.GET.get("q", "")
        return context


class SearchView(HeliportLoginRequiredMixin, TemplateView):
    """Handle searching objects in the HELIPORT core app."""

    #: HTML template rendered by this View
    template_name = "core/search/search.html"

    def get_context_data(self, **kwargs):
        """Query all objects matching "q" parameter."""
        context = super().get_context_data(**kwargs)
        user = self.request.user.heliportuser
        query = self.request.GET.get("q", "")
        context["query"] = query

        project_results = set()
        group_results = set()
        user_results = set()
        login_info_results = set()

        for word in query.split():
            num = int(word) if word.isdecimal() else -1

            project_results.update(
                Project.objects.filter(
                    (Q(owner=user) | Q(co_owners=user))
                    & Q(deleted__isnull=True)
                    & (
                        Q(project_id=num)
                        | Q(hzdr_id__icontains=word)
                        | Q(label__icontains=word)
                        | Q(owner__display_name__icontains=word)
                        | Q(co_owners__display_name__icontains=word)
                        | Q(description__icontains=word)
                        | Q(persistent_id__icontains=word)
                        | Q(group__display_name=word)
                    )
                )
            )

            group_results.update(Group.objects.filter(display_name__icontains=word))

            user_results.update(
                HeliportUser.objects.filter(
                    Q(display_name__icontains=word) | Q(user_id=num)
                )
            )

            login_info_results.update(
                LoginInfo.objects.filter(
                    Q(user=user)
                    & (
                        Q(username__icontains=word)
                        | Q(machine__icontains=word)
                        | Q(name__icontains=word)
                        | Q(type__icontains=word)
                    )
                )
            )

        context["project_results"] = project_results
        context["group_results"] = group_results
        context["user_results"] = user_results
        context["login_info_results"] = login_info_results
        return context


class MaintenanceMessageView(View):
    """Handles maintenance message specifiable in settings."""

    def get(self, request, **kwargs):
        """Return json object with the "message" and "level" (is translated to color of the banner)"""
        return JsonResponse(
            {
                "message": settings.HELIPORT_CORE_MAINTENANCE_MESSAGE,
                "level": settings.HELIPORT_CORE_MAINTENANCE_MESSAGE_LEVEL,
            }
        )


class ManualUserDownloadView(HeliportLoginRequiredMixin, View):
    """Handle manually starting user import task. See issue #449."""

    def get(self, request, **kwargs):
        result = heliport.core.tasks.import_user_data()
        return HttpResponse(f"OK. {result} users imported")


##########################################
#               REST API                 #
##########################################


class LoginNamedTokenView(knox.views.LoginView):
    """
    Customized `LoginView <https://github.com/James1345/django-rest-knox/blob/3a1bc584f9691f4bc19d8a04a98c68c293be9ca6/knox/views.py#L15>`_
    for named HELIPORT API tokens.
    """

    def get_token_ttl(self):
        """Allow "expiry" parameter of the form YYYY-MM-DD. Empty string means no expiry."""
        expiry = self.request.POST.get("expiry")
        if expiry is None:
            return super().get_token_ttl()
        if expiry == "":
            return None

        try:
            expire_time = datetime.strptime(expiry, "%Y-%m-%d")
        except ValueError:
            return RestResponse(
                {
                    "error": '"expiry" has invalid format; supported format is YYYY-MM-DD'
                },
                status=status.HTTP_400_BAD_REQUEST,
            )
        return timezone.make_aware(expire_time) - timezone.now()

    def get_post_response_data(self, request, token, instance):
        """Put "hash" into response for displaying token."""
        result = super().get_post_response_data(request, token, instance)
        result["hash"] = instance.pk

        # In the next version of knox there is a `create_token()` method where this code would fit better
        # https://github.com/James1345/django-rest-knox/blob/develop/knox/views.py#L41
        name = self.request.POST.get("name")
        if name:
            NamedToken.objects.create(auth_token=instance, name=name)

        return result


class LogoutTokenView(APIView):
    """
    Handle deleting token. Similar to knox `LogoutView <https://github.com/James1345/django-rest-knox/blob/3a1bc584f9691f4bc19d8a04a98c68c293be9ca6/knox/views.py#L78>`_
    but can delete token that is not used to authenticate the current request.
    """

    permission_classes = [rest_permissions.IsAuthenticated]

    def post(self, request, *args, **kwargs):
        """Delete token specified by "pk" parameter."""
        token = get_object_or_404(
            knox_models.AuthToken, pk=self.request.POST.get("pk"), user=request.user
        )
        token.delete()
        user_logged_out.send(
            sender=request.user.__class__, request=request, user=request.user
        )
        return RestResponse(None, status=status.HTTP_204_NO_CONTENT)


class DeleteAllTokensView(APIView):
    """
    Handle deleting all tokens. Similar to knox `LogoutAllView <https://github.com/James1345/django-rest-knox/blob/3a1bc584f9691f4bc19d8a04a98c68c293be9ca6/knox/views.py#L89>`_
    but uses the default authentication classes.
    """

    permission_classes = [rest_permissions.IsAuthenticated]

    def post(self, request, *args, **kwargs):
        request.user.auth_token_set.all().delete()
        user_logged_out.send(
            sender=request.user.__class__, request=request, user=request.user
        )
        return RestResponse(None, status=status.HTTP_204_NO_CONTENT)


class HeliportModelViewSet(viewsets.ModelViewSet):
    """Base class for API ViewSets in HELIPORT for DigitalObjects."""

    #: Make sure user is authenticated and is allowed to access the DigitalObject
    permission_classes = [rest_permissions.IsAuthenticated, HeliportObjectPermission]

    def perform_destroy(self, instance):
        """Deleting objects by setting `deleted` to the current time."""
        instance.deleted = timezone.now()
        instance.save()


class ProjectViewSet(HeliportModelViewSet):
    """
    Showing the most general HELIPORT project properties.
    Request more detailed information by appending urls/ to the url
    Go to a specific project by appending its id/ to the url
    For more information on how to authenticate look in HELIPORT (user > settings)
    """

    serializer_class = ProjectSerializer
    permission_classes = [rest_permissions.IsAuthenticated, HeliportProjectPermission]
    filter_backends = [
        rest_filters.SearchFilter,
        django_filters.rest_framework.DjangoFilterBackend,
    ]
    search_fields = ["label", "description", "owner__display_name"]
    filterset_fields = ["group", "owner", "persistent_id"]

    def get_queryset(self):
        """Get projects of current user that are not deleted."""
        user = self.request.user.heliportuser
        return projects_of(user).filter(deleted__isnull=True).distinct()

    @action(detail=False)
    def urls(self, request):
        """Action to get urls to apis of HELIPORT apps when appending /urls/ to the url of this view."""
        api_roots = [
            f"{ request.scheme }://{ request.get_host() }{url}"
            for url in get_api_roots()
        ]
        return RestResponse(api_roots)


class TokenViewSet(viewsets.ModelViewSet):
    """Showing HELIPORT user tokens."""

    permission_classes = [rest_permissions.IsAuthenticated]
    filter_backends = [
        rest_filters.SearchFilter,
        django_filters.rest_framework.DjangoFilterBackend,
    ]
    filterset_fields = ["login_info_id"]
    search_fields = ["name"]
    serializer_class = TokenSerializer

    def get_queryset(self):
        """Get tokens (stored in heliport as LoginInfos, not HELIPORT API tokens) of current user"""
        user = self.request.user.heliportuser
        return LoginInfo.objects.filter(user=user, type=LoginInfo.LoginTypes.TOKEN)


class UserViewSet(viewsets.ModelViewSet):
    """Showing HELIPORT user model."""

    permission_classes = [rest_permissions.IsAuthenticated, StaffOrReadOnly]
    filterset_fields = ["user_id", "authentication_backend_id", "display_name"]

    queryset = HeliportUser.objects.all()
    serializer_class = UserSerializer


class GroupViewSet(viewsets.ModelViewSet):
    """Showing user groups."""

    permission_classes = [rest_permissions.IsAuthenticated, StaffOrReadOnly]

    queryset = Group.objects.all()
    serializer_class = GroupSerializer


class ContributionViewSet(viewsets.ModelViewSet):
    """Showing contributions to projects of the current user."""

    permission_class = [rest_permissions.IsAuthenticated, StaffOrReadOnly]
    serializer_class = ContributionSerializer

    def get_queryset(self):
        """return contributions to projects of the user that are not deleted."""
        user = self.request.user.heliportuser
        projects = projects_of(user).filter(deleted__isnull=True)
        # note __in=projects computes no intermediate result in python
        # but __in=list(project) respects the fact that Project is a subclass of DigitalObject
        return Contribution.objects.filter(
            contribution_to__in=list(projects)
        ).distinct()


class DynamicBrowsableAPIRenderer(BrowsableAPIRenderer):
    """Render dynamic API. Contains javascript and css needed for :class:`DigitalObjectAPI`."""

    #: HTML template rendered by this View
    template = "core/api/dynamic_api_base.html"


class DigitalObjectAPI(HeliportModelViewSet):
    """API for all digital objects in HELIPORT."""

    #: :class:`DigitalObjectAPISerializer` contains the code that makes this general api possible.
    serializer_class = DigitalObjectAPISerializer
    #: This api allows access to all DigitalObjects. Permissions (read/write) are respected per attribute.
    queryset = DigitalObject.objects.all()
    #: Even without login public attributes are accessible.
    permission_classes = [rest_permissions.IsAuthenticatedOrReadOnly]
    #: Replace normal html renderer with :class:`DynamicBrowsableAPIRenderer` for some extra javascript and css.
    renderer_classes = [
        DynamicBrowsableAPIRenderer if c is BrowsableAPIRenderer else c
        for c in api_settings.DEFAULT_RENDERER_CLASSES
    ]

    def __init__(self, *args, **kwargs):
        """Initialize json-ld context to collect property namespaces."""
        super().__init__(*args, **kwargs)
        self.ld_context = LdContext()

    def finalize_response(self, request, response, *args, **kwargs):
        """Add json-ld context to response to make this api more interoperable."""
        response = super().finalize_response(request, response, *args, **kwargs)
        if isinstance(response.data, OrderedDict):
            response.data["@context"] = self.ld_context.get_context_dict()
            response.data.move_to_end("@context", last=False)
        elif isinstance(response.data, dict):
            response.data = dict(
                **{"@context": self.ld_context.get_context_dict()}, **response.data
            )
        return response


# the open api schema view read by redoc
SchemaView = get_schema_view(
    title="HELIPORT",
    description="API for all things …",
    version=heliport_version_string,
)


class RedocView(TemplateView):
    """Render the redoc API docu."""

    #: HTML template rendered by this View
    template_name = "core/api/redoc.html"
