# SPDX-FileCopyrightText: 2022 Helmholtz-Zentrum Dresden-Rossendorf (HZDR)
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
import pathlib
from stat import S_ISDIR
from urllib.parse import quote

from django.conf import settings
from django.contrib import messages
from django.db.models import Q
from django.http import FileResponse, JsonResponse
from django.shortcuts import get_object_or_404, redirect, render, reverse
from django.utils import timezone
from django.views.generic import CreateView, DetailView, TemplateView, UpdateView

from heliport.core.custom_views import (
    HeliportLoginRequiredMixin,
    HeliportObjectMixin,
    HeliportProjectMixin,
)
from heliport.core.models import Project, Vocabulary, assert_relation
from heliport.core.permissions import HeliportObjectPermission, is_object_member
from heliport.core.views import HeliportModelViewSet

from .models import DataSource
from .serializers import DataSourceSerializer

logger = logging.getLogger(__name__)


class DataSourceView(HeliportProjectMixin, CreateView):
    model = DataSource
    template_name = "data_source/data_source.html"
    fields = ["uri", "description", "label", "login"]

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        project = context["project"]
        context["vocab"] = Vocabulary()
        context["sources"] = [
            {
                "data_source_id": source.data_source_id,
                "digital_object_id": source.digital_object_id,
                "protocol": source.uri.split(":")[0],
                "label": source.label,
                "description": source.description,
                "uri": source.uri,
                "browser_supported_uri": source.browser_supported_uri,
                "owner": (source.login is None)
                or (self.request.user.heliportuser == source.login.user),
            }
            for source in DataSource.objects.filter(
                projects=project, deleted__isnull=True, generated_inside__isnull=True
            )
        ]
        user = self.request.user.heliportuser
        context["protocols"] = DataSource.PROTOCOL_CHOICES
        context["logins"] = user.logininfo_set.all()
        context["update"] = False
        return context

    def post(self, request, *args, **kwargs):
        data_source_id = request.POST.get("remove")
        if data_source_id:
            data_source = get_object_or_404(DataSource, pk=data_source_id)
            data_source.deleted = timezone.now()
            data_source.save()
            return redirect(reverse("data_source:list", kwargs=self.kwargs))
        else:
            return super().post(request, *args, **kwargs)

    def form_valid(self, form):
        protocol = self.request.POST["protocol"]
        server_adress = self.request.POST["server_adress"]

        if protocol == "http" and not form.instance.uri.startswith("http://"):
            form.instance.uri = "http://" + form.instance.uri
        elif protocol == "https" and not form.instance.uri.startswith("https://"):
            form.instance.uri = "https://" + form.instance.uri
        elif protocol == "file" and not form.instance.uri.startswith("file://"):
            form.instance.uri = "file://" + form.instance.uri
        elif protocol == "ftp":
            if not form.instance.uri.startswith("/") and not server_adress.endswith(
                "/"
            ):
                form.instance.uri = "/" + form.instance.uri
            form.instance.uri = server_adress + form.instance.uri
            if not form.instance.uri.startswith("ftp://"):
                form.instance.uri = "ftp://" + form.instance.uri
        elif protocol == "ftps":
            if not form.instance.uri.startswith("/") and not server_adress.endswith(
                "/"
            ):
                form.instance.uri = "/" + form.instance.uri
            form.instance.uri = server_adress + form.instance.uri
            if not form.instance.uri.startswith("ftps://"):
                form.instance.uri = "ftps://" + form.instance.uri
        elif protocol == "ssh":
            form.instance.uri = "ssh://" + form.instance.uri
        else:
            if "://" not in form.instance.uri:
                form.instance.uri = "://" + form.instance.uri

        form.instance.category_str = settings.DATA_SOURCE_NAMESPACE

        form.instance.save()
        project = get_object_or_404(Project, pk=self.kwargs["project"])
        form.instance.projects.add(project)
        form.instance.register_handle_if_applicable(self.request)

        return super().form_valid(form)

    def get_success_url(self):
        return reverse("data_source:list", kwargs=self.kwargs)


class DataSourceUpdateView(HeliportObjectMixin, UpdateView):
    model = DataSource
    template_name = "data_source/data_source.html"
    fields = ["uri", "description", "label", "login"]

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        project = context["project"]
        context["vocab"] = Vocabulary()
        context["sources"] = [
            {
                "data_source_id": source.data_source_id,
                "digital_object_id": source.digital_object_id,
                "protocol": source.uri.split(":")[0],
                "label": source.label,
                "description": source.description,
                "uri": source.uri,
                "browser_supported_uri": source.browser_supported_uri,
                "owner": (source.login is None)
                or (self.request.user.heliportuser == source.login.user),
            }
            for source in DataSource.objects.filter(
                projects=project, deleted__isnull=True, generated_inside__isnull=True
            )
        ]
        protocol = self.request.GET.get("protocol", self.object.uri.split("://")[0])
        user = self.request.user.heliportuser

        if protocol == ("ftp" or "ftps"):
            context["server_adress"] = self.object.uri.split("://", 1)[1].split("/", 1)[
                0
            ]
            context["uri"] = self.object.uri.split("://", 1)[1].split("/", 1)[1]
        else:
            context["uri"] = self.object.uri.split("://", 1)[1]

        context["protocol"] = protocol
        context["protocols"] = DataSource.PROTOCOL_CHOICES
        context["logins"] = user.logininfo_set.all()
        context["login"] = self.object.login
        context["update"] = True
        return context

    def form_valid(self, form):
        if (self.object.login is None) or (
            self.request.user.heliportuser == self.object.login.user
        ):
            protocol = self.request.POST["protocol"]
            server_adress = self.request.POST["server_adress"]

            if protocol == "http" and not form.instance.uri.startswith("http://"):
                form.instance.uri = "http://" + form.instance.uri
            elif protocol == "https" and not form.instance.uri.startswith("https://"):
                form.instance.uri = "https://" + form.instance.uri
            elif protocol == "file" and not form.instance.uri.startswith("file://"):
                form.instance.uri = "file://" + form.instance.uri
            elif protocol == "ftp":
                if not form.instance.uri.startswith("/") and not server_adress.endswith(
                    "/"
                ):
                    form.instance.uri = "/" + form.instance.uri
                form.instance.uri = server_adress + form.instance.uri
                if not form.instance.uri.startswith("ftp://"):
                    form.instance.uri = "ftp://" + form.instance.uri
            elif protocol == "ftps":
                if not form.instance.uri.startswith("/") and not server_adress.endswith(
                    "/"
                ):
                    form.instance.uri = "/" + form.instance.uri
                form.instance.uri = server_adress + form.instance.uri
                if not form.instance.uri.startswith("ftps://"):
                    form.instance.uri = "ftps://" + form.instance.uri
            elif protocol == "ssh":
                form.instance.uri = "ssh://" + form.instance.uri
            else:
                if "://" not in form.instance.uri:
                    form.instance.uri = "://" + form.instance.uri

        return super().form_valid(form)

    def get_success_url(self):
        return reverse("data_source:list", kwargs={"project": self.kwargs["project"]})


class SSHDataSourceOpenView(HeliportObjectMixin, DetailView):
    model = DataSource

    def get(self, request, *args, **kwargs):
        self.object = self.get_object()
        data_source = self.object
        project = get_object_or_404(Project, pk=self.kwargs["project"])
        login = data_source.login
        source_path = pathlib.Path(data_source.path)
        relative_path = pathlib.Path("")
        if "path" in request.GET:
            relative_path = pathlib.Path(request.GET["path"])
            assert not any(
                n.startswith(".") for n in relative_path.parts
            ), "access of hidden directory"

        path = source_path / relative_path
        assert (
            source_path == path or source_path in path.parents
        ), "access of non sub directory"

        logger.debug(f"opening ssh connection datasource={data_source.pk}")
        ssh_connection = login.build_command_executor()
        error_message = ssh_connection.connect_error_message(
            f' - try reconnecting the "{login.name}" ssh connection'
        )
        if error_message is not None:
            messages.error(request, error_message)
            return redirect("data_source:list", project=project.pk)
        file = ssh_connection.file(str(path))

        file_stat = file.stat()
        if hasattr(file_stat, "st_mode") and S_ISDIR(file_stat.st_mode):
            logger.info(f"opening ssh datasource {data_source.pk} (Directory)")
            context = super().get_context_data()
            dir_content_files = ssh_connection.list_dir(str(path))
            dir_content = [
                {
                    "name": s.filename,
                    "path": relative_path / s.filename,
                    "quoted_path": quote(str(relative_path / s.filename)),
                    "digital_object": data_source_by_path(
                        project, path / s.filename, data_source
                    ),
                    "size": hasattr(s, "st_size") and self.format_size(s.st_size),
                    "mtime": hasattr(s, "st_mtime") and self.format_time(s.st_mtime),
                    "is_dir": hasattr(s, "st_mode") and S_ISDIR(s.st_mode),
                }
                for s in dir_content_files
                if not s.filename.startswith(".")
            ]
            context["dir_content"] = sorted(
                dir_content, key=lambda n: (not n["is_dir"], n["name"].lower())
            )
            data_source_to_remove = DataSource.objects.filter(
                projects=project,
                uri__startswith=f"ssh://{path}/",
                deleted__isnull=True,
                generated_inside=data_source,
            )
            for f in dir_content_files:
                data_source_to_remove = data_source_to_remove.exclude(
                    uri__startswith=f"ssh://{path / f.filename}"
                )
            for data_source in data_source_to_remove:
                if path not in pathlib.Path(data_source.path).parents:
                    continue
                logger.debug(f"removing data source {data_source.uri}")
                data_source.deleted = timezone.now()
                data_source.save()

            parent_paths = []
            for p in reversed(relative_path.parents):
                if p != pathlib.Path():
                    parent_paths.append({"name": p.name, "path": str(p)})

            context["parent_paths"] = parent_paths
            context["directory_name"] = path.name
            context["has_relative_path"] = path != pathlib.Path(data_source.path)
            context["vocab"] = Vocabulary()
            return render(request, "data_source/directory.html", context)

        old_close = file.close

        def new_close():
            old_close()
            ssh_connection.close()
            logger.debug(f"closing ssh connection datasource={data_source.pk}")

        file.close = new_close
        response = FileResponse(file, filename=path.name, as_attachment=True)
        file_stat = file.stat()
        if hasattr(file_stat, "st_size"):
            logger.info(
                f"opening ssh datasource {data_source.pk} ({self.format_size(file_stat.st_size)})"
            )
            response["Content-Length"] = file_stat.st_size
        else:
            logger.info(f"opening ssh datasource {data_source.pk}")
        return response

    @staticmethod
    def format_size(size):
        for unit in ["Bytes", "KiB", "MiB", "GiB", "TiB", "PiB"]:
            if size < 1024.0 or unit == "PiB":
                break
            size /= 1024.0
        return f"{size: .0f} {unit}"

    @staticmethod
    def format_time(mtime):
        return timezone.datetime.fromtimestamp(mtime)


def data_source_by_path(project, path, parent_data_source):
    uri = f"ssh://{path}"
    return (
        DataSource.objects.filter(
            uri=uri,
            login=parent_data_source.login,
            projects=project,
            deleted__isnull=True,
        )
        .order_by("pk")
        .first()
    )


class SubDataSourceCreateView(HeliportObjectMixin, DetailView):
    model = DataSource

    def post(self, request, *args, **kwargs):
        project = get_object_or_404(Project, pk=self.kwargs["project"])
        data_source = self.get_object()
        assert data_source.uri.startswith("ssh://")
        login = data_source.login
        path = data_source.sub_path(request.GET.get("path"))
        existing_data_source = data_source_by_path(project, path, data_source)
        if existing_data_source is None:
            new_data_source = DataSource(
                login=login,
                uri=f"ssh://{path}",
                label=path.name,
                generated_inside=data_source,
            )
            new_data_source.category_str = settings.DATA_SOURCE_NAMESPACE
            new_data_source.save()
            new_data_source.projects.add(*data_source.projects.all())
        else:
            new_data_source = existing_data_source
        assert_relation(data_source, Vocabulary().has_part, new_data_source)
        return JsonResponse(
            {
                "pk": new_data_source.pk,
                "digital_object_id": new_data_source.digital_object_id,
                "uri": new_data_source.uri,
                "label": new_data_source.label,
            }
        )


class SearchView(HeliportLoginRequiredMixin, TemplateView):
    template_name = "data_source/search.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        user = self.request.user.heliportuser
        query = self.request.GET.get("q", "")
        context["query"] = query

        data_source_results = set()

        for word in query.split():
            num = int(word) if word.isdecimal() else -1

            data_source_results.update(
                DataSource.objects.filter(
                    is_object_member(user)
                    & Q(projects__deleted__isnull=True, deleted__isnull=True)
                    & (
                        Q(data_source_id=num)
                        | Q(uri__icontains=word)
                        | Q(description__icontains=word)
                        | Q(label__icontains=word)
                        | Q(login__machine__icontains=word)
                        | Q(login__username__icontains=word)
                        | Q(projects__label__icontains=word)
                    )
                )
            )

        context["data_source_results"] = data_source_results
        return context


##########################################
#               REST API                 #
##########################################


class DataSourceViewSet(HeliportModelViewSet):
    """Data Source"""

    serializer_class = DataSourceSerializer
    filterset_fields = ["data_source_id", "label", "uri", "persistent_id", "projects"]

    def get_queryset(self):
        user = self.request.user.heliportuser
        return DataSource.objects.filter(
            is_object_member(user) & Q(deleted__isnull=True)
        ).distinct()
