# SPDX-FileCopyrightText: 2022 Helmholtz-Zentrum Dresden-Rossendorf (HZDR)
#
# SPDX-License-Identifier: GPL-3.0-or-later

import json
import logging

from django.http import HttpResponse, HttpResponseBadRequest, HttpResponseNotFound
from django.shortcuts import get_object_or_404, redirect
from django.urls import reverse
from django.utils.dateparse import parse_datetime
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_POST
from django.views.generic import TemplateView, View
from pyunicore.client import Job

from heliport.core.custom_views import HeliportProjectMixin
from heliport.core.models import Project

from .decorators import require_content_type
from .mixins import UnicoreMixin, UnicoreServerAuthenticationRequiredMixin
from .models import UnicoreResource

logger = logging.getLogger(__name__)


class Overview(UnicoreMixin, TemplateView):
    """Overview of the user's UNICORE resources.

    When this page is viewed, the UNICORE jobs and storages are saved in the database
    as :class:`unicore.models.UnicoreResource` objects.
    """

    template_name = "unicore/overview.html"

    def get_context_data(self, **kwargs):
        """Create method context.

        For more info see `Django docs <https://docs.djangoproject.com/en/4.0/ref/class-based-views/mixins-simple/#django.views.generic.base.ContextMixin.get_context_data>`_.
        """
        context = super().get_context_data(**kwargs)

        if self.unicore_client is not None:
            job_list = self.unicore_client.get_jobs()
            storage_list = self.unicore_client.get_storages()

            for job in job_list:
                to_remove = []
                for storage in storage_list:
                    if (job.job_id + "-uspace") in storage.resource_url:
                        job.properties["storage"] = storage
                        to_remove.append(storage)
                for storage in to_remove:
                    storage_list.remove(storage)

            for job in job_list:
                job_obj, _ = UnicoreResource.objects.get_or_create(
                    resource_id=job.job_id, resource_url=job.resource_url
                )
                job.properties["submissionTimeDateTime"] = parse_datetime(
                    job.properties["submissionTime"]
                )

            job_list.reverse()
            context["unicore_job_list"] = job_list
            context["unicore_storage_list"] = storage_list

        return context


class ProjectJobsView(UnicoreMixin, HeliportProjectMixin, TemplateView):
    """List of the UNICORE jobs that belong to a project.

    :param unicore_job_tag_project: Jobs with this tag are automatically associated with
        this project.
    :type unicore_job_tag_project: str
    """

    template_name = "unicore/project-jobs.html"

    def __init__(self):
        super().__init__()
        self.unicore_job_tag_project = None

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self.unicore_job_tag_project = (
            f"{self.unicore_job_tag_prefix}-{kwargs['project']}"
        )

    def get_context_data(self, **kwargs):
        """Create method context.

        For more info see `Django docs <https://docs.djangoproject.com/en/4.0/ref/class-based-views/mixins-simple/#django.views.generic.base.ContextMixin.get_context_data>`_.
        """
        context = super().get_context_data(**kwargs)

        if self.unicore_client is not None:
            job_list = self.unicore_client.get_jobs(tags=[self.unicore_job_tag_project])
            project = Project.objects.get(pk=kwargs["project"])

            for job in job_list:
                # TODO: Create these objects in Overview but for all projects the user
                #  is in.
                job_obj, _ = UnicoreResource.objects.get_or_create(
                    resource_id=job.job_id, resource_url=job.resource_url
                )
                if self.unicore_job_tag_project in job.properties["tags"]:
                    job_obj.heliport_project = project
                    job_obj.save()

                job.properties["submissionTimeDateTime"] = parse_datetime(
                    job.properties["submissionTime"]
                )

            job_list.reverse()
            context["unicore_job_list"] = job_list

        context["unicore_job_tag_project"] = self.unicore_job_tag_project
        return context


class JobActionView(UnicoreMixin, View):
    """View that allows the user to perform some actions on a UNICORE job.

    :param allowed_actions: The actions that can be performed on a job
    :type allowed_actions: List[str]
    """

    def __init__(self):
        super().__init__()
        self.allowed_actions = ["abort", "delete", "restart"]

    def get(self, request, job_id, action):
        """Method called on GET request.

        For more info see `Django docs <https://docs.djangoproject.com/en/4.0/ref/class-based-views/base/#django.views.generic.base.View.dispatch>`_.
        """
        if action not in self.allowed_actions:
            return HttpResponseNotFound()
        logger.debug(f"Action {action!r} requested for UNICORE job {job_id}")
        job = get_object_or_404(UnicoreResource, resource_id=job_id)
        if self.unicore_transport is not None:
            getattr(Job(self.unicore_transport, job.resource_url), action)()
        job.delete()  # delete the job from DB in case it was deleted in UNICORE
        if "project" in request.GET:
            return redirect(
                reverse("unicore:project-jobs", args=(request.GET["project"],))
            )
        return redirect(reverse("unicore:overview"))


class TestJobView(UnicoreMixin, View):
    """View that, on behalf of the user, sends a job to the cluster.

    This is ment for testing purposes. The job runs on the login node and executes this
    code:

    .. code-block:: bash

       echo Hello World

    If the URL parameter ``project`` is passed along, the job will be tagged so that it
    can be found in the appropriate project.
    """

    def get(self, request):
        """Method called on GET request.

        For more info see `Django docs <https://docs.djangoproject.com/en/4.0/ref/class-based-views/base/#django.views.generic.base.View.dispatch>`_.
        """

        project = request.GET.get("project")

        if self.unicore_client is not None:
            job_description = {
                "Name": "HELIPORT Test Job",
                # Emailing doesn't work in "INTERACTIVE" jobs
                # "User email": request.user.heliportuser.auth_user.email,
                "Notification": request.build_absolute_uri(reverse("unicore:callback")),
                "Job type": "INTERACTIVE",  # run on login node
                "Executable": "echo",
                "Arguments": ["Hello", "World"],
                "Tags": ["test"],
            }
            if project is not None:
                job_description["Tags"].append(
                    f"{self.unicore_job_tag_prefix}-{project}"
                )
            self.unicore_client.new_job(job_description)

        if project is not None:
            return redirect(reverse("unicore:project-jobs", args=(project,)))
        return redirect(reverse("unicore:overview"))


@method_decorator(csrf_exempt, name="dispatch")
@method_decorator(require_POST, name="dispatch")
@method_decorator(require_content_type("application/json"), name="dispatch")
class CallbackView(UnicoreServerAuthenticationRequiredMixin, View):
    """Can be called by UNICORE to notify HELIPORT about a job's status."""

    def post(self, request, *args, **kwargs):
        """Method called on POST request.

        For more info see `Django docs <https://docs.djangoproject.com/en/4.0/ref/class-based-views/base/#django.views.generic.base.View.dispatch>`_.
        """

        try:
            data = json.loads(request.body)
            logger.info(
                f"Notification from UNICORE server {self.unicore_server_dn} "
                f"for {self.unicore_user_dn}: {data}"
            )
        except json.decoder.JSONDecodeError:
            logger.debug(f"{request.get_full_path()} called with malformed data")
            return HttpResponseBadRequest()
        return HttpResponse(status=204)
