# SPDX-FileCopyrightText: 2022 Helmholtz-Zentrum Dresden-Rossendorf (HZDR)
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""
Custom collections similar to builtin collections like `dict` or `list` but with additional or different functionality.
"""
import typing
from collections import UserDict, defaultdict


class SingleAssignDict(dict):
    """A `dict` that has an option to disallow reassigning keys."""

    def __init__(self, allow_reassign_key=False):
        super().__init__()
        self.allow_reassign_key = allow_reassign_key

    def __setitem__(self, key, value):
        if not self.allow_reassign_key:
            assert key not in self, f"re-registering {key} with {value} forbidden"
        super().__setitem__(key, value)


class DecoratorDict(SingleAssignDict):
    """
    A `dict` that provides :func:`register` decorator and can be used like this::

        function_collection = DecoratorDict()

        @function_collection.register("a nice function")
        def a_nice_function():
            pass

        function_collection["a nice function"]()
    """

    def __init__(self, allow_reassign_key=False):
        super().__init__(allow_reassign_key)

    def register(self, key):
        """Decorator to assign a function to `key`"""

        def register_key(value):
            self[key] = value
            return value

        return register_key


class DecoratorIndex(SingleAssignDict):
    """
    A `dict` that provides :func:`register` decorator for classes. It can be used like this::

        numbered_classes = DecoratorIndex("x")

        @numbered_classes.register
        class OneClass:
            x = 1

        one_instance = numbered_classes[1]()
    """

    def __init__(self, field, allow_reassign_key=False):
        super().__init__(allow_reassign_key)
        self.field = field

    def register(self, value):
        """Decorator to assign a class to the key specified by `field`."""
        key = getattr(value, self.field)
        self[key] = value
        return value


class DecoratorList(list):
    """
    A `list` that provides :func:`register` decorator to append a function or class to the list::

        my_functions = DecoratorList()

        @my_functions.register
        def f():
            pass

        my_functions[0]()
    """

    register = list.append


class IndexedSet:
    """
    ``IndexedSet`` behaves similar to pythons ``set`` however it provides additional query functions::

        Fruit = namedtuple("Fruit", "name color")
        fruits = IndexedSet()
        fruits.add(Fruit("Apple", "red"))
        fruits.add(Fruit("Walnut", "brown"))
        fruits.get_by(color="red")  # returns Fruit("Apple", "red")

    All members in an ``IndexedSet`` must be hashable.
    All methods (except :meth:`build_index`) return in constant amortized time
    for a fixed number of key combinations used in queries.
    """

    def __init__(self):
        self.data = set()
        self.indexes = dict()

    def add(self, item: typing.Hashable):
        """
        Add ``item`` to this ``IndexedSet``.
        ``add`` takes time linear in the number of indexes (see :meth:`filter_by`).
        """
        for index_name, index in self.indexes.items():
            index_entry = tuple(getattr(item, n) for n in index_name)
            index[index_entry].add(item)
        self.data.add(item)

    def discard(self, item: typing.Hashable):
        """
        Remove ``item`` from this ``IndexedSet`` if present.
        ``discard`` takes time linear in the number of indexes (see :meth:`filter_by`).
        """
        for index_name, index in self.indexes.items():
            index_entry = tuple(getattr(item, n) for n in index_name)
            index[index_entry].discard(item)
        self.data.discard(item)

    def remove(self, item):
        """
        Remove ``item`` from this ``IndexedSet``.
        ``remove`` takes time linear in the number of indexes (see :meth:`filter_by`).

        :raise KeyError: if not present
        """
        if item not in self:
            raise KeyError(item)
        return self.discard(item)

    def __iter__(self):
        return iter(self.data)

    def __contains__(self, item):
        return item in self.data

    def filter_by(self, **conditions) -> set:
        """
        Returns all items in this ``IndexedSet``, that meet the specified conditions, as a ``set``.
        Normally this function runs in constant time,
        however the first time, some combination of keys is used, an index is created in linear time.
        """

        index_name = tuple(sorted(conditions))
        index_entry = tuple(conditions[n] for n in index_name)
        index = self.indexes.get(index_name)
        if index is None:
            index = self.build_index(index_name)
            self.indexes[index_name] = index

        return index[index_entry]

    def get_by(self, **conditions):
        """
        Similar to :meth:`find_by` but returns only one item.

        :raise AssertionError: if there is no or more than one item matching the ``conditions``
        """
        result = self.filter_by(**conditions)
        assert len(result) == 1
        return next(iter(result))

    def get_any_or_none(self, **conditions):
        """
        Similar to :meth:`find_by` but returns only one item.
        If no items match ``conditions``, ``None`` is returned
        """
        result = self.filter_by(**conditions)
        if result:
            return next(iter(result))
        else:
            return None

    def build_index(self, index_name: typing.Tuple[str]):
        """
        Used by :meth:`find_by` to build indices.

        :param index_name: Note that order matters. Please sort the tuple before passing into this function.
        """
        index = defaultdict(set)
        for value in self:
            index_entry = tuple(getattr(value, n) for n in index_name)
            index[index_entry].add(value)
        return index


class HashableDictWrapper(UserDict):
    """
    Normally ``dict`` s are not hashable in python because they are mutable for the default notion of equality.
    Two ``HashableDictWrapper`` s are equal only if they are the same object
    (like classes that don't implement ``__eq__``).
    Under this notion dictionaries are immutable and can be hashable.
    """

    def __init__(self, the_dict):
        super().__init__()
        self.data = the_dict

    def __eq__(self, other):
        if not isinstance(other, HashableDictWrapper):
            return False
        return self.data is other.data

    def __hash__(self):
        return hash(id(self))
