# SPDX-FileCopyrightText: 2022 Helmholtz-Zentrum Dresden-Rossendorf (HZDR)
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Collection of functions that work with colors."""

import random


def pick_color() -> tuple:
    """Return color from uniform distribution as tuple of three numbers in ``range(255)``"""
    return tuple(random.choice(range(255)) for i in range(3))


def color_dist(c1, c2):
    """
    Compute distance of colors based on some heuristic.

    :param c1: A color represented by three values e.g. (1, 2, 3)
    :param c2: A color represented by three values e.g. (1, 2, 3)
    :return: Score how different ``c1`` and ``c2`` are. Larger score means more different.
    """
    r1, g1, b1 = c1
    r2, g2, b2 = c2
    return 0.3 * (r1 - r2) ** 2 + 0.59 * (g1 - g2) ** 2 + 0.11 * (b1 - b2) ** 2


def color_dist_array(c1, colors):
    """
    Compute minimum distance of ``c1`` to any color in ``colors``.
    Input and output format is the same as in :func:`color_dist`.
    """
    return min((color_dist(c1, c2) for c2 in colors), default=0)


def pick_color_not_in(colors):
    """
    Pick random color. Colors more different from all colores in ``colors`` are preferred.
    For color representation see :func:`pick_color` and :func:`color_dist` functions.
    """
    color = pick_color()
    current_dist = color_dist_array(color, colors)
    for i in range(200):
        new_color = pick_color()
        new_dist = color_dist_array(new_color, colors)
        if new_dist > current_dist:
            current_dist = new_dist
            color = new_color
    return color


def pick_foreground_color(background_html_color: str) -> str:
    """
    Return a color visible on some ``background_html_color``.
    Colores are represented like in HTML e.g. ``#FFF`` (different length, lowercase or without ``#`` is also supported)
    """
    if not background_html_color or " " in background_html_color:
        # invalid colors are rendered as white so we chose black
        return "#000000"
    color = background_html_color
    if color.startswith("#"):
        color = color[1:]
    if len(color) not in [3, 6, 8]:
        return "#000000"
    if len(color) > 3:
        color = color[::2]
    light_count = sum(c >= "9" for c in color)
    if light_count >= 2 or any(c.lower() > "c" for c in color[:2]):
        return "#000000"
    else:
        return "#FFFFFF"
