# SPDX-FileCopyrightText: 2022 Helmholtz-Zentrum Dresden-Rossendorf (HZDR)
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Helper functions to work with strings"""

from datetime import datetime
from typing import List, Union


def remove_prefix(
    chars: str, prefix_options: Union[str, List[str]], raise_on_error=False
) -> str:
    """
    Remove a prefix (potentially from a list of prefixes) from ``chars``.
    Note that this function should be used instead of builtin ``str.removeprefix`` as long
    as we support versions that don't have this function.

    :param chars: The string before removal of the prefix.
    :param prefix_options: The first matching prefix option is removed.
        Supplying a string is equivalent to a list with one element.
    :param raise_on_error: If ``False`` nothing happens when no prefix matches
    :raise ValueError: If ``raise_on_error`` is True and no prefix matches
    :return: ``chars`` but without the prefix
    """

    if isinstance(prefix_options, str):
        prefix_list = [prefix_options]
    else:
        prefix_list = prefix_options

    for prefix in prefix_list:
        if chars.startswith(prefix):
            return chars[len(prefix) :]

    if raise_on_error:
        raise ValueError(f"{chars} does not start with {', '.join(prefix_list)}")
    else:
        return chars


def parse_date_time(string_date):
    if not isinstance(string_date, str):
        return None
    if string_date in {"now", "today"}:
        return datetime.now()
    for fmt in [
        "%b %d, %Y",
        "%b %d, %Y %H:%M",
        "%Y-%m-%d",
        "%Y-%m-%d %H-%M",
        "%d-%b-%y",
        "%m/%d/%Y",
        "%d. %m. %Y",
    ]:
        try:
            return datetime.strptime(string_date, fmt)
        except ValueError:
            continue
    return None
