# SPDX-FileCopyrightText: 2022 Helmholtz-Zentrum Dresden-Rossendorf (HZDR)
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""The overall configuration of the HELIPORT Django application."""

from pathlib import Path

import environ
from celery.schedules import crontab
from django.core.exceptions import ImproperlyConfigured

BASE_DIR = Path(__file__).resolve().parent.parent

env = environ.Env()
env.read_env()

# Run HELIPORT in development mode. This setting has security implications and should
# only be set to true for development!
DEVELOPMENT = env.bool("DEVELOPMENT", default=False)


# =============================== USERS, AUTHENTICATION ===============================

LDAP_ENABLED = env.bool("LDAP_ENABLED", default=False)

# User sync in heliport.core
if LDAP_ENABLED:
    LDAP_SERVER_URL = env.str("LDAP_SERVER_URL")
    LDAP_BIND_USER_DN = env.str("LDAP_BIND_USER_DN")
    LDAP_BIND_PASSWORD = env.str("LDAP_BIND_PASSWORD")
    LDAP_LOGIN_NAME_SEARCH = env.str("LDAP_LOGIN_NAME_SEARCH")
    if not "%(user)s" in LDAP_LOGIN_NAME_SEARCH:
        raise ImproperlyConfigured(
            "LDAP_LOGIN_NAME_SEARCH must contain the placeholder `$(user)s`."
        )
    LDAP_USER_BASE_DN = env.str("LDAP_USER_BASE_DN")
    LDAP_USER_ID_ATTRIBUTE = env.str("LDAP_USER_ID_ATTRIBUTE")
    LDAP_USER_DISPLAY_NAME_ATTRIBUTE = env.str("LDAP_USER_DISPLAY_NAME_ATTRIBUTE")
    LDAP_USER_ATTRIBUTE_MAP = env.dict(
        "LDAP_USER_ATTRIBUTE_MAP",
        default={
            "first_name": "givenName",
            "last_name": "sn",
            "email": "mail",
        },
    )
    LDAP_USER_OBJECT_CLASS = env.str("LDAP_USER_OBJECT_CLASS", default="person")
    LDAP_GROUP_BASE_DN = env.str("LDAP_GROUP_BASE_DN")
    LDAP_GROUP_ID_ATTRIBUTE = env.str("LDAP_GROUP_ID_ATTRIBUTE")
    LDAP_GROUP_DISPLAY_NAME_ATTRIBUTE = env.str("LDAP_GROUP_DISPLAY_NAME_ATTRIBUTE")
    LDAP_GROUP_MEMBER_ATTRIBUTE = env.str("LDAP_GROUP_MEMBER_ATTRIBUTE")
    LDAP_GROUP_FILTER = env.str("LDAP_GROUP_FILTER", default=None)
    LDAP_GROUP_OBJECT_CLASS = env.str("LDAP_GROUP_OBJECT_CLASS", default="groupOfNames")

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
]

AUTHENTICATION_BACKEND_USED = "ldap" if LDAP_ENABLED else "basic"
AUTHENTICATION_BACKENDS = ["django.contrib.auth.backends.ModelBackend"]
if AUTHENTICATION_BACKEND_USED == "ldap":
    AUTHENTICATION_BACKENDS.insert(0, "django_auth_ldap.backend.LDAPBackend")

# Settings for django-auth-ldap
if AUTHENTICATION_BACKEND_USED == "ldap":
    import ldap
    from django_auth_ldap.config import LDAPSearch, NestedMemberDNGroupType

    AUTH_LDAP_SERVER_URI = LDAP_SERVER_URL
    AUTH_LDAP_BIND_DN = LDAP_BIND_USER_DN
    AUTH_LDAP_BIND_PASSWORD = LDAP_BIND_PASSWORD
    AUTH_LDAP_USER_SEARCH = LDAPSearch(
        LDAP_USER_BASE_DN,
        ldap.SCOPE_SUBTREE,
        LDAP_LOGIN_NAME_SEARCH,
    )
    AUTH_LDAP_GROUP_SEARCH = LDAPSearch(
        LDAP_GROUP_BASE_DN, ldap.SCOPE_SUBTREE, LDAP_GROUP_FILTER
    )
    AUTH_LDAP_GROUP_TYPE = NestedMemberDNGroupType(
        LDAP_GROUP_MEMBER_ATTRIBUTE, name_attr=LDAP_GROUP_DISPLAY_NAME_ATTRIBUTE
    )
    AUTH_LDAP_MIRROR_GROUPS = False
    AUTH_LDAP_USER_ATTR_MAP = LDAP_USER_ATTRIBUTE_MAP


# ================================== HELIPORT SETUP ===================================

HELIPORT_GATE_ENABLED = env.bool("HELIPORT_GATE_ENABLED", default=False)
if HELIPORT_GATE_ENABLED:
    if not LDAP_ENABLED:
        raise ImproperlyConfigured("HELIPORT_GATE_ENABLED requires LDAP_ENABLED")

HELIPORT_UNICORE_ENABLED = env.bool("HELIPORT_UNICORE_ENABLED", default=False)

HELIPORT_SYSTEM_APPS = [
    "heliport.core",
    "heliport.version_control",
    "heliport.data_management_plan",
    "heliport.documentation",
    "heliport.data_source",
    "heliport.archive",
    "heliport.publication",
    "heliport.cwl_execution",
    "heliport.digital_objects",
    "heliport.sharelatex",
    "heliport.tables",
]

if HELIPORT_GATE_ENABLED:
    HELIPORT_SYSTEM_APPS.append("heliport.gate_connection")

if HELIPORT_UNICORE_ENABLED:
    HELIPORT_SYSTEM_APPS.append("heliport.unicore")

HELIPORT_MIDDLEWARE = [
    "heliport.core.middleware.never_cache_middleware.DisableClientSideCachingMiddleware",
]

# Project graph
HELIPORT_GRAPH = {
    "Project": [
        "project_configuration",
        "gate_connection",
        "timeline",
        "object_graph",
        "tags",
        "subprojects",
    ],
    "Systems": [
        "version_control",
        "data_management_plan",
        "documentation",
        "digital_objects",
        "sharelatex",
        "unicore",
    ],
    "Resources": [
        "data_source",
        "arbitrary_cwl",
        "tool",
        "workflow",
        "tables",
        "images",
    ],
    "Automation": [
        "job",
        "unicore_project_jobs",  # this is probably not the best place
    ],
    "Results": ["archive", "publication"],
}

# Dependencies between HELIPORT apps
HELIPORT_DEPENDENCIES = {
    "version_control": ["project_configuration"],
    "data_management_plan": ["project_configuration"],
    "documentation": ["project_configuration"],
    "data_source": ["project_configuration"],
    "archive": ["project_configuration"],
    "publication": ["project_configuration"],
    "job": ["project_configuration"],
    "tool": ["project_configuration"],
    "workflow": ["project_configuration"],
    "arbitrary_cwl": ["project_configuration"],
    "digital_objects": ["project_configuration"],
    "sharelatex": ["project_configuration"],
    "tables": ["project_configuration"],
    "images": ["project_configuration"],
    "timeline": ["project_configuration"],
    "object_graph": ["project_configuration"],
    "tags": ["project_configuration"],
    "subprojects": ["project_configuration"],
    "unicore": ["project_configuration"],
    "unicore_project_jobs": ["project_configuration"],
}

# Namespaces
# TODO: Maybe these can be moved somewhere so that each app (that needs it) has it
#  automatically? I.e. a HELIPORT-specific AppConfig/AppConf that all apps inherit from?
ARCHIVE_NAMESPACE = "HELIPORT/archive"
EXECUTABLE_ARBITRARY_CWL_NAMESPACE = "HELIPORT/cwl_file"
EXECUTABLE_WORKFLOW_NAMESPACE = "HELIPORT/workflow"
EXECUTABLE_TOOL_NAMESPACE = "HELIPORT/tool"
JOB_NAMESPACE = "HELIPORT/job"
DATA_MANAGEMENT_PLAN_NAMESPACE = "HELIPORT/data_management_plan"
DATA_SOURCE_NAMESPACE = "HELIPORT/data_source"
DOCUMENTATION_NAMESPACE = "HELIPORT/documentation"
PUBLICATION_NAMESPACE = "HELIPORT/publication"
VERSION_CONTROL_NAMESPACE = "HELIPORT/version_control"
PROPOSAL_NAMESPACE = "HELIPORT/proposal"
SHARE_LATEX_NAMESPACE = "HELIPORT/share_latex"
TABLE_NAMESPACE = "HELIPORT/table"
IMAGE_NAMESPACE = "HELIPORT/image"
GRAPH_VISUALIZATION_NAMESPACE = "HELIPORT/graph_visualization"
TAG_NAMESPACE = "HELIPORT/tag"
PROPERTY_NAMESPACE = "property"
CLASS_NAMESPACE = "class"
DEFAULT_OBJECT_NAMESPACE = "object"


# =================================== DJANGO SETUP ====================================

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "rest_framework",  # for REST API
    "knox",  # for using Token authentication
    "django_filters",  # for adding query arguments to rest url
    "django_celery_results",
    "health_check",
    "health_check.db",
    "health_check.cache",
    "health_check.storage",
] + HELIPORT_SYSTEM_APPS

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
] + HELIPORT_MIDDLEWARE

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

STATIC_URL = "/static/"
STATIC_ROOT = "static"
STATICFILES_DIRS = []

DATABASES = {
    "default": env.db_url("DB_URL", default=f"sqlite:///{BASE_DIR}/heliport.sqlite3")
}

CACHES = {
    "default": {
        "BACKEND": "django.core.cache.backends.locmem.LocMemCache",
        "LOCATION": "heliport-default-cache",
    },
    "image_cache": {
        "BACKEND": "django.core.cache.backends.filebased.FileBasedCache",
        "LOCATION": "/var/tmp/heliport_image_cache",
    },
}

LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "handlers": {
        "file": {
            "class": "logging.FileHandler",
            "filename": BASE_DIR / "heliport.log"
            if DEVELOPMENT
            else "/var/log/heliport/heliport.log",
            "formatter": "default",
        },
        "console": {
            "class": "logging.StreamHandler",
            "formatter": "default",
        },
    },
    "loggers": {
        "root": {
            "handlers": ["file"] + (["console"] if DEVELOPMENT else []),
            "level": "DEBUG" if DEVELOPMENT else "INFO",
        },
    },
    "formatters": {
        "default": {
            "format": "{asctime} - {name} - {levelname} :-: {message}",
            "datefmt": "%m/%d/%Y %I:%M:%S %p",
            "style": "{",
        }
    },
}

WSGI_APPLICATION = "heliport_config.wsgi.application"
ROOT_URLCONF = "heliport_config.urls"

# Security settings
DEBUG = DEVELOPMENT
SECRET_KEY = env.str("SECRET_KEY")
ALLOWED_HOSTS = env.list("ALLOWED_HOSTS", default=[]) + ["127.0.0.1", "localhost"]
CSRF_COOKIE_SECURE = not DEVELOPMENT
SESSION_COOKIE_SECURE = not DEVELOPMENT
SECURE_SSL_REDIRECT = not DEVELOPMENT
SESSION_EXPIRE_AT_BROWSER_CLOSE = True

LANGUAGES = [("en", "English")]
USE_I18N = False
USE_TZ = True
TIME_ZONE = env.str("TIME_ZONE", default="Europe/Berlin")

# For image upload
DATA_UPLOAD_MAX_MEMORY_SIZE = 36 * 1024 * 1024  # 36 MiB

# Settings for django-rest-framework
REST_FRAMEWORK = {
    "DEFAULT_PAGINATION_CLASS": "rest_framework.pagination.LimitOffsetPagination",
    "PAGE_SIZE": 10,
    "AUTO_REFRESH": True,
    "MIN_REFRESH_INTERVAL": 60,  # seconds
    "DEFAULT_AUTHENTICATION_CLASSES": [
        "knox.auth.TokenAuthentication",
        "rest_framework.authentication.SessionAuthentication",
    ],
    "DEFAULT_PERMISSION_CLASSES": [
        "rest_framework.permissions.IsAuthenticated",
    ],
    "DEFAULT_FILTER_BACKENDS": [
        "django_filters.rest_framework.DjangoFilterBackend",
    ],
}


# =================================== CELERY SETUP ====================================

CELERY_BROKER_URL = env.str("CELERY_BROKER_URL")
CELERY_RESULT_BACKEND = env.str("CELERY_RESULT_BACKEND", default="django-db")
# Deprecated by Celery but still used by django_celery_results
CELERY_CACHE_BACKEND = "django-cache" if CELERY_RESULT_BACKEND == "django-db" else None

CELERY_RESULT_EXPIRES = 5 * 24 * 60 * 60  # 5 days

CELERY_SERIALIZER = "json"
CELERY_RESULT_SERIALIZER = "json"
CELERY_TASK_SERIALIZER = "json"
CELERY_ACCEPT_CONTENT = ["json"]
CELERY_TIMEZONE = TIME_ZONE

CELERY_BEAT_SCHEDULE = {
    "periodic_job_execution": {
        "task": "heliport.cwl_execution.tasks.run_periodic_jobs",
        "schedule": crontab(minute=50, hour=2),
    },
}

if LDAP_ENABLED:
    CELERY_BEAT_SCHEDULE["daily_ldap_sync"] = {
        "task": "heliport.core.tasks.import_user_data",
        "schedule": crontab(minute=10, hour=2),
    }

if HELIPORT_GATE_ENABLED:
    CELERY_BEAT_SCHEDULE["daily_gate_sync"] = {
        "task": "heliport.gate_connection.tasks.update_gate_projects",
        "schedule": crontab(minute=30, hour=2),
    }
